import pytest
import numpy as np
import SimpleITK as sitk
from pyRadPlan.core.np2sitk import (
    linear_indices_to_sitk_mask,
    linear_indices_to_grid_coordinates,
    linear_indices_to_image_coordinates,
    sitk_mask_to_linear_indices,
)
from pyRadPlan.core._grids import Grid


@pytest.fixture
def ref_image() -> sitk.Image:
    # Create a reference SimpleITK image
    ref_image = sitk.Image(3, 4, 5, sitk.sitkUInt8)
    ref_image.SetSpacing((1.0, 2.0, 3.0))
    ref_image.SetOrigin((0.0, -20.0, 5.0))
    ref_image.SetDirection(np.eye(3).flatten())
    return ref_image


@pytest.fixture
def grid(ref_image: sitk.Image) -> Grid:
    # Create a reference Grid object
    return Grid.from_sitk_image(ref_image)


def test_linear_indices_to_sitk_mask_sitk_order(ref_image: sitk.Image):
    # Create a reference SimpleITK image

    indices = np.array([0, 4, 25])
    indices_unraveled = np.unravel_index(indices, ref_image.GetSize())

    # Convert linear indices to SimpleITK mask
    mask = linear_indices_to_sitk_mask(indices, ref_image, order="sitk")

    # Check the mask properties
    mask_array = sitk.GetArrayViewFromImage(mask)
    assert mask_array.shape == ref_image.GetSize()[::-1]

    for i in range(len(indices)):
        assert (
            mask.GetPixel(
                int(indices_unraveled[0][i]),
                int(indices_unraveled[1][i]),
                int(indices_unraveled[2][i]),
            )
            == 1
        )

    assert mask.GetPixel(1, 0, 0) == 0


def test_linear_indices_to_sitk_mask_numpy_order(ref_image: sitk.Image):
    # Create a reference SimpleITK image

    indices = np.array([0, 4, 25])
    indices_unraveled = np.unravel_index(indices, ref_image.GetSize(), order="F")

    # Convert linear indices to SimpleITK mask
    mask = linear_indices_to_sitk_mask(indices, ref_image, order="numpy")

    # Check the mask properties
    mask_array = sitk.GetArrayViewFromImage(mask)
    assert mask_array.shape == ref_image.GetSize()[::-1]

    for i in range(len(indices)):
        assert (
            mask.GetPixel(
                int(indices_unraveled[0][i]),
                int(indices_unraveled[1][i]),
                int(indices_unraveled[2][i]),
            )
            == 1
        )

    assert np.all(mask_array.flat[indices] == 1)


def test_linear_indices_to_sitk_mask_invalid_order(ref_image: sitk.Image):
    # dummy indices
    indices = np.array([0, 1])

    # Check for invalid order
    with pytest.raises(ValueError):
        mask = linear_indices_to_sitk_mask(indices, ref_image, order="invalid")


def test_sitk_mask_to_linear_indices_sitk_order(ref_image: sitk.Image):
    indices = np.array([0, 4, 25])

    # Convert linear indices to SimpleITK mask
    mask = linear_indices_to_sitk_mask(indices, ref_image, order="sitk")
    new_indices = sitk_mask_to_linear_indices(mask, order="sitk")
    assert np.all(new_indices == indices)


def test_sitk_mask_to_linear_indices_numpy_order(ref_image: sitk.Image):
    indices = np.array([0, 4, 25])

    # Convert linear indices to SimpleITK mask
    mask = linear_indices_to_sitk_mask(indices, ref_image, order="numpy")
    new_indices = sitk_mask_to_linear_indices(mask, order="numpy")
    assert np.all(new_indices == indices)


def test_sitk_mask_to_linear_indices_invalid_order(ref_image: sitk.Image):
    # dummy indices
    indices = np.array([0, 1])

    # Convert linear indices to SimpleITK mask
    mask = linear_indices_to_sitk_mask(indices, ref_image, order="numpy")

    # Check for invalid order
    with pytest.raises(ValueError):
        new_indices = sitk_mask_to_linear_indices(mask, order="invalid")


def test_linear_indices_to_grid_coordinates_numpy_order(grid: Grid, ref_image: sitk.Image):
    indices = np.array([0, 4, 25])
    indices_unraveled = np.unravel_index(indices, grid.dimensions, order="F")
    coordinates = linear_indices_to_grid_coordinates(indices, grid, index_type="numpy")

    for i in range(len(indices)):
        ix_tuple = (
            int(indices_unraveled[0][i]),
            int(indices_unraveled[1][i]),
            int(indices_unraveled[2][i]),
        )
        assert np.all(
            np.isclose(ref_image.TransformIndexToPhysicalPoint(ix_tuple), coordinates[i, :])
        )


def test_linear_indices_to_grid_coordinates_sitk_order(grid: Grid, ref_image: sitk.Image):
    indices = np.array([0, 4, 25])
    indices_unraveled = np.unravel_index(indices, grid.dimensions)
    coordinates = linear_indices_to_grid_coordinates(indices, grid, index_type="sitk")

    for i in range(len(indices)):
        ix_tuple = (
            int(indices_unraveled[0][i]),
            int(indices_unraveled[1][i]),
            int(indices_unraveled[2][i]),
        )
        assert np.all(
            np.isclose(ref_image.TransformIndexToPhysicalPoint(ix_tuple), coordinates[i, :])
        )


def test_linear_indices_to_grid_coordinates_invalid_order(grid: Grid):
    indices = np.array([0, 1])

    with pytest.raises(ValueError):
        coordinates = linear_indices_to_grid_coordinates(indices, grid, index_type="invalid")


def test_linear_indices_to_image_coordinates(grid: Grid, ref_image: sitk.Image):
    indices = np.array([0, 4, 25])
    assert np.all(
        np.isclose(
            linear_indices_to_grid_coordinates(indices, grid),
            linear_indices_to_image_coordinates(indices, ref_image),
        )
    )

    assert np.all(
        np.isclose(
            linear_indices_to_grid_coordinates(indices, grid, index_type="numpy"),
            linear_indices_to_image_coordinates(indices, ref_image, index_type="numpy"),
        )
    )

    with pytest.raises(ValueError):
        coordinates = linear_indices_to_image_coordinates(indices, ref_image, index_type="invalid")



import numpy as np

from pyRadPlan.optimization.solvers import get_solver, OptimizerSciPy, SolverBase


def test_get_solver_scipy():
    solver = get_solver("scipy")
    assert isinstance(solver, OptimizerSciPy)
    assert isinstance(solver, SolverBase)
    assert solver.short_name == "scipy"
    assert solver.method == "L-BFGS-B"


def test_simple_problem_scipy():
    solver = get_solver("scipy")

    # Define the problem
    def objective(x):
        return x[0] ** 2 + x[1] ** 2

    def gradient(x):
        return [2 * x[0], 2 * x[1]]

    solver.objective = objective
    solver.gradient = gradient

    # Initial guess
    x0 = [1.0, 1.0]

    # Solve
    result = solver.solve(x0)

    assert np.all(np.isclose(result[0], 0.0))



"""Read and write mat-files."""

from os import PathLike
from scipy import io
import pymatreader


def load(path2file: PathLike) -> dict[str]:
    """Load a .mat file and returns the data as a dictionary."""

    try:
        matrad_patient = pymatreader.read_mat(filename=path2file)
    except NotImplementedError:
        matrad_patient = io.loadmat(
            file_name=path2file, mat_dtype=True, squeeze_me=True, struct_as_record=True
        )
    except Exception as exep:
        raise ValueError(f"Could not load the .mat file: {path2file}") from exep

    return matrad_patient


def save(path2file, the_dict: dict[str]):
    """
    Save a dictionary as a .mat file.

    Parameters
    ----------
    path2file : str
        Path to the file.
    the_dict : dict[str]
        Dictionary to be saved.
    """

    io.savemat(file_name=path2file, mdict=the_dict)



"""Factory methods to manage available solver implementations."""

import warnings
import logging
from typing import Union, Type
from ._base_solvers import SolverBase

SOLVERS = {}

logger = logging.getLogger(__name__)


def register_solver(solver_cls: Type[SolverBase]) -> None:
    """
    Register a new solver.

    Parameters
    ----------
    solver_cls : type
        A Dose Solver class.
    """
    if not issubclass(solver_cls, SolverBase):
        raise ValueError("Solver must be a subclass of SolverBase.")

    if solver_cls.short_name is None:
        raise ValueError("Solver must have a 'short_name' attribute.")

    if solver_cls.name is None:
        raise ValueError("Solver must have a 'name' attribute.")

    solver_name = solver_cls.short_name
    if solver_name in SOLVERS:
        warnings.warn(f"Solver '{solver_name}' is already registered.")
    else:
        SOLVERS[solver_name] = solver_cls


def get_available_solvers() -> dict[str, Type[SolverBase]]:
    """
    Get a list of available solvers based on the plan.

    Returns
    -------
    list
        A list of available solvers.
    """
    return SOLVERS


def get_solver(solver_desc: Union[str, dict, SolverBase]):
    """
    Returns a solver instance based on a descriptive parameter.

    Parameters
    ----------
    solver_desc : Union[str, dict, SolverBase]
        A string with the solver name, a dictionary with the solver configuration or a solver
        instance

    Returns
    -------
    SolverBase
        A solver instance
    """
    if isinstance(solver_desc, str):
        solver = SOLVERS[solver_desc]()
    elif isinstance(solver_desc, dict):
        raise NotImplementedError("Solver configuration from dictionary not implemented yet.")
    elif isinstance(solver_desc, SolverBase):
        solver = solver_desc
    else:
        raise ValueError(f"Invalid solver description: {solver_desc}")

    return solver



from typing import cast
import pytest
import sys

if sys.version_info < (3, 10):
    import importlib_resources as resources  # Backport for older versions
else:
    from importlib import resources  # Standard from Python 3.9+

import SimpleITK as sitk
import numpy as np

from pyRadPlan.io import load_patient
from pyRadPlan.core import np2sitk
from pyRadPlan.ct import validate_ct, CT
from pyRadPlan.cst import validate_cst, StructureSet


@pytest.fixture
def tg119():
    path = resources.files("pyRadPlan.data.phantoms")
    ct, cst = load_patient(path.joinpath("TG119.mat"))
    return {"ct": ct, "cst": cst}


def test_linear_index_to_world(tg119):
    ct = cast(CT, tg119["ct"])
    cst = cast(StructureSet, tg119["cst"])

    ct_array = sitk.GetArrayViewFromImage(ct.cube_hu)

    np_index = cst.vois[0].indices_numpy
    sitk_index = cst.vois[0].indices
    np_index_raveled = np.unravel_index(np_index, ct_array.shape)
    np_index_raveled = np.array(
        [np_index_raveled[0][0], np_index_raveled[1][0], np_index_raveled[2][0]]
    ).astype(np.int_)

    sitk_index = tuple([v.item() for v in np_index_raveled[::-1]])

    sitk_point = np.array(ct.cube_hu.TransformIndexToPhysicalPoint(sitk_index))

    np_coord = np2sitk.linear_indices_to_image_coordinates(
        indices=np_index[0], image=ct.cube_hu, index_type="numpy"
    )

    assert np.isclose(sitk_point, np_coord).all()



from ._engine import MatRadEngine
import os


class MatRadEngineOctave(MatRadEngine):
    @property
    def engine(self):
        return self.__engine

    @property
    def matRadVersionStr(self):
        return self.__matRadVersionStr

    @property
    def engineVersionStr(self):
        return self.__engineVersionStr

    def __init__(
        self,
        matRadPath="./matRad",
        octave_exec="C:/Program Files/GNU Octave/Octave-8.3.0/mingw64/bin/octave-cli.exe",
    ):
        import oct2py

        self.matRadPath = matRadPath
        self.octave_exec = octave_exec

        oct2pyInit = oct2py.Oct2Py()
        self.__engineVersionStr = oct2pyInit.version()

        # from oct2py import octave

        self.__engine = oct2pyInit
        self.__engine.addpath(self.matRadPath)
        self.__engine.matRad_rc
        self.__matRadVersionStr = self.engine.matRad_version()
        # print(self.matRadVersionStr)

    @property
    def octave_exec(self):
        return self._octave_exec

    @octave_exec.setter
    def octave_exec(self, path):
        # Check if executable
        executable = os.access(path, os.X_OK)
        if not executable:
            raise ValueError("Octave executable not found")

        os.environ["OCTAVE_EXECUTABLE"] = path
        self._octave_exec = path



"""Defines a class representing a single ray.

The ray is pointing from the beam source to a position in the patient.
"""

import functools
from typing import Any, Union, Optional
import numpy as np
from pydantic import (
    create_model,
    Field,
    model_validator,
    field_validator,
    ValidationInfo,
    ValidatorFunctionWrapHandler,
    field_serializer,
    SerializationInfo,
    SerializerFunctionWrapHandler,
    computed_field,
    ValidationError,
)
from numpydantic import NDArray, Shape
from pyRadPlan.util.helpers import dl2ld, ld2dl
from pyRadPlan.core import PyRadPlanBaseModel
from pyRadPlan.stf._beamlet import Beamlet


class Ray(PyRadPlanBaseModel):
    """
    A class representing a single ray.

    This class extends PyRadPlanBaseModel (pydantic) and provides functionality to
    handle the ray information, including properties like ray positions, energy, focus index, etc.

    Attributes
    ----------
    ray_pos_bev : np.ndarray
        The ray positions in BEV coordinates.
    ray_pos : np.ndarray
        The ray positions in (x, y, z) coordinates.
    target_point_bev : np.ndarray
        The target point in BEV coordinates.
    target_point : np.ndarray
        The target point in (x, y, z) coordinates.
    beamlets : list[Beamlet]
        The beamlets in the ray.
    """

    beamlets: list[Beamlet]

    ray_pos_bev: NDArray[Shape["3"], np.float64] = Field(alias="rayPos_bev")
    ray_pos: NDArray[Shape["3"], np.float64]

    target_point: Optional[NDArray[Shape["3"], np.float64]] = Field(default=None)
    target_point_bev: Optional[NDArray[Shape["3"], np.float64]] = Field(
        alias="targetPoint_bev", default=None
    )

    @field_validator("ray_pos_bev", "target_point_bev", "ray_pos", "target_point", mode="wrap")
    @classmethod
    def validate_nparray_dtype(
        cls, v: Any, handler: ValidatorFunctionWrapHandler, info: ValidationInfo
    ) -> NDArray[Shape["3"], np.float64]:
        """Validate / convert arrays to have floating point values."""

        if v is not None:
            if not isinstance(v, np.ndarray):
                v = np.array(v, dtype=float)
            if not np.issubdtype(v.dtype, np.floating):
                v = v.astype(float)
            v = v.reshape((3,))
        return handler(v, info)

    # Custom validator to convert list to np.ndarray
    @model_validator(mode="wrap")
    @classmethod
    def sanitize_beamlet_structure(
        cls, data: Any, handler: ValidatorFunctionWrapHandler, info: ValidationInfo
    ) -> Any:
        """
        Sanitize the beamlet structure in the ray.

        It may be structured differently (e.g. when coming from matRad.).
        """
        # if isinstance(data, list):
        #     for i in range(len(data)):
        #         if isinstance(data[i], int) or isinstance(data[i], float):
        #             data[i] = [float(data[i])]
        #         if isinstance(data[i], list):
        #             data[i] = np.array(data[i])

        try:
            return handler(data, info)
        except ValidationError:
            if isinstance(data, dict):
                # We obtain some meta information about the Beamlet model
                beamlet_fields = Beamlet.model_fields
                beamlet_field_aliases = {
                    field.validation_alias: key for key, field in beamlet_fields.items()
                }

                # Beamlets may be structured differently in the ray, so we need to extract them
                # into a dictionary containing a list of values (or other dicts) for each beamlet
                # property
                beamlet_subdict = {}
                to_remove = []

                for key in data:
                    if key in beamlet_fields.keys() or key in beamlet_field_aliases.keys():
                        if key in beamlet_field_aliases.keys():
                            update_key = beamlet_field_aliases[key]
                        else:
                            update_key = key
                        value = data[key]

                        if isinstance(value, dict):
                            try:
                                value = dl2ld(
                                    value, type_check=True
                                )  # dict_of_lists to list_of_dicts
                            except TypeError:
                                value = [value]
                                # This is an exception for the case where the beamlet is a single
                                # beamlet

                        # now, if it is not a list, we make it a list
                        if not isinstance(value, list) and not isinstance(value, np.ndarray):
                            value = [value]

                        beamlet_subdict.update({update_key: value})

                        to_remove.append(key)

                # Sanitze data to not have beamlet properties as arrays in the ray
                for key in to_remove:
                    data.pop(key)

                # correct indexing for focus_ix:
                if "focus_ix" in beamlet_subdict:
                    beamlet_subdict["focus_ix"] = [ix - 1 for ix in beamlet_subdict["focus_ix"]]

                try:
                    beamlets = dl2ld(beamlet_subdict, type_check=False)
                except TypeError as exc:
                    raise TypeError(f"Beamlet information not consistent in Ray: {exc}") from exc

                data["beamlets"] = beamlets

            return handler(data)
        except Exception as exc:
            raise exc

    @field_serializer("beamlets", mode="wrap")
    def custom_beamlets_serializer(
        self, v: list[Beamlet], handler: SerializerFunctionWrapHandler, info: SerializationInfo
    ) -> Union[dict[str, Any], list]:
        context = info.context
        if context and context.get("matRad") == "mat-file":
            beamlets_dump = [beamlet.to_matrad(context=context["matRad"]) for beamlet in v]

            # Convert the list of dictionaries to a dictionary of lists and return
            beamlets_dump = ld2dl(beamlets_dump, type_check=False)

            for field in beamlets_dump:
                first_element = beamlets_dump[field][0]
                if isinstance(first_element, dict):
                    field_dump = ld2dl(beamlets_dump[field], type_check=False)
                    beamlets_dump[field] = np.rec.fromarrays(
                        field_dump.values(), names=list(field_dump)
                    )
            return beamlets_dump
        return handler(v, info)

    def to_matrad(self, context: Union[str, dict] = "mat-file") -> Any:
        """Serialize rays for matRad structure."""

        model_dump = super().to_matrad(context=context)

        for key in model_dump["beamlets"]:
            model_dump[key] = model_dump["beamlets"][key]

        model_dump.pop("beamlets")
        return model_dump

    @classmethod
    def create_matrad_helper_model(cls):
        """Create a helper model for matRad serialization."""

        def get_property_list(self: Ray, name: str):
            return [getattr(beamlet, name) for beamlet in self.beamlets]

        beamlet_fields = Beamlet.model_fields
        prop_lambdas = {}
        for field in beamlet_fields:
            prop_lambdas[field] = computed_field(
                functools.partial(get_property_list, name=field),
                return_type=list[beamlet_fields[field].annotation],
                alias=beamlet_fields[field].serialization_alias,
            )

        return create_model("RayMatRadHelper", __base__=Ray, **prop_lambdas)



"""Squared Deviation Objective."""

from typing import Annotated
from pydantic import Field

from numba import njit

from ._objective import Objective, ParameterMetadata

# %% Class definition


class SquaredDeviation(Objective):
    """
    Squared Deviation (least-squares) objective.

    Attributes
    ----------
    d_ref : float
        dose reference value
    """

    name = "Squared Deviation"

    d_ref: Annotated[float, Field(default=60.0, ge=0.0), ParameterMetadata(kind="reference")]

    def compute_objective(self, values):
        return _compute_objective(values, self.d_ref)

    def compute_gradient(self, values):
        return _compute_gradient(values, self.d_ref)


@njit
def _compute_objective(dose, d_ref):
    deviation = dose - d_ref

    return (deviation @ deviation) / len(dose)


@njit
def _compute_gradient(dose, d_ref):
    return 2.0 * (dose - d_ref) / len(dose)



"""Factory methods to manage available objective implementations."""

import warnings
import logging
from typing import Union, Type
from ._objective import Objective

__matrad_name_map__ = {
    "DoseObjectives.matRad_SquaredDeviation": "Squared Deviation",
    "DoseObjectives.matRad_SquaredUnderdosing": "Squared Underdosing",
    "DoseObjectives.matRad_SquaredOverdosing": "Squared Overdosing",
    "DoseObjectives.matRad_MeanDose": "Mean Dose",
    "DoseObjectives.matRad_EUD": "EUD",
    "DoseObjectives.matRad_MinDVH": "MinDVH",
    "DoseObjectives.matRad_MaxDVH": "MaxDVH",
}

OBJECTIVES = {}

logger = logging.getLogger(__name__)


def register_objective(obj_cls: Type[Objective]) -> None:
    """
    Register a new objective.

    Parameters
    ----------
    obj_cls : type
        An Objective class.
    """
    if not issubclass(obj_cls, Objective):
        raise ValueError("Objective must be a subclass of Objective.")

    if obj_cls.name is None:
        raise ValueError("Objective must have a 'name' attribute.")

    obj_name = obj_cls.name
    if obj_name in OBJECTIVES:
        warnings.warn(f"Objective '{obj_name}' is already registered.")
    else:
        OBJECTIVES[obj_name] = obj_cls


def get_available_objectives() -> dict[str, Type[Objective]]:
    """
    Get a list of available objectives.

    Returns
    -------
    list
        A list of available objectives.
    """
    return OBJECTIVES


def get_objective(objective_desc: Union[str, dict, Objective]):
    """
    Returns a objective instance based on a descriptive parameter.

    Parameters
    ----------
    objective_desc : Union[str, dict, Objective]
        A string with the objective name, a dictionary with the objective configuration or a
        objective instance

    Returns
    -------
    Objective
        A objective instance
    """
    if isinstance(objective_desc, str):
        objective = OBJECTIVES[objective_desc]()
    elif isinstance(objective_desc, dict):
        if "name" not in objective_desc:
            logger.debug("Objective not found, trying matRad-like objective.")
            if "className" not in objective_desc:
                raise ValueError(f"Invalid objective description: {objective_desc}")
            objective_name = __matrad_name_map__.get(objective_desc["className"], None)
            if objective_name is None:
                raise ValueError(f"Invalid objective description: {objective_desc}")
        else:
            objective_name = objective_desc["name"]

        objective_model = OBJECTIVES[objective_name]
        objective = objective_model.model_validate(objective_desc)
    elif isinstance(objective_desc, Objective):
        objective = objective_desc
    else:
        raise ValueError(f"Invalid objective description: {objective_desc}")

    return objective



"""
Sceanrio Models.

Used for uncertainty quantification, robust optimization and
robustness analysis.
"""

from typing import Union, Optional, Type
from pyRadPlan.ct import CT, validate_ct
from ._base import ScenarioModel
from ._nominal import NominalScenario
import logging
from copy import deepcopy

logger = logging.getLogger(__name__)


def available_scenario_models() -> list[str]:
    """Return a list of available scenario models."""

    return ["nomScen"]


def create_scenario_model(
    model_def: Union[ScenarioModel, Union[str, dict]] = "nomScen",
    ct: Optional[Union[CT, dict]] = None,
) -> ScenarioModel:
    """
    Return a scenario model object.

    Parameters
    ----------
    model_name : str
        The name of the scenario model.

    Returns
    -------
    ScenarioModel
        The scenario model object.
    """

    if isinstance(ct, dict):
        ct = validate_ct(ct)

    if isinstance(model_def, str):
        model_class = _get_model_class_by_name(model_def)
        return model_class(ct)

    if isinstance(model_def, ScenarioModel):
        return model_def

    if isinstance(model_def, dict):
        model_name = model_def.pop("model", None)
        if model_name is None:
            raise ValueError("Scenario model name not provided in the model definition")

        model_class = _get_model_class_by_name(model_name)
        return model_class(ct, **model_def)

    raise ValueError("Invalid scenario model definition")


def _get_model_class_by_name(model_name: str) -> Type[ScenarioModel]:
    """Return the scenario model class by name."""

    if model_name == "nomScen":
        return NominalScenario
    if model_name == "wcScen":
        raise NotImplementedError("Worst Case Scenarios are not implemented yet")
    if model_name == "impScen":
        raise NotImplementedError("Gridded Importance Weighted Scenarios are not implemented yet")
    if model_name == "rndScen":
        raise NotImplementedError("Random Scenarios are not implemented yet")
    raise ValueError(f"Unknown scenario model: {model_name}")


def validate_scenario_model(
    model_def: Union[ScenarioModel, Union[str, dict]], ct: Union[CT, dict] = None
) -> ScenarioModel:
    """
    Validate a scenario model input and returns a scenario model object.

    Parameters
    ----------
    model_def : Union[ScenarioModel, Union[str, dict]]
        The scenario model object.

    Returns
    -------
    ScenarioModel
        The scenario model object.
    """
    model_def = deepcopy(model_def)
    if "ctScenProb" in model_def:
        if all(a > 0 for a, b in model_def["ctScenProb"]):
            model_def["ctScenProb"] = [(a - 1, b) for (a, b) in model_def["ctScenProb"]]
            logger.info(
                "Assuming scenario model from matRad. Converting 1-based indexing to 0-based "
                "indexing..."
            )
        else:
            logger.info(
                "CamelCase was used, but assuming 0-based indexing, since some values are less "
                "than 1. No index conversion applied"
            )

    return create_scenario_model(model_def, ct)


__all__ = ["ScenarioModel", "NominalScenario", "create_scenario_model"]



"""
Implementation of a pencil beam dose calculation engine for photon beams
based on the Singular-value
decomposition (SVD) method by Bortfeld.
"""

from typing import TypedDict, Literal, Any, cast, Callable
import logging
import random

import numpy as np
from scipy import fft
from scipy.interpolate import RegularGridInterpolator

from pyRadPlan.plan import PhotonPlan

# from pyRadPlan.stf import Beam
from pyRadPlan.machines import PhotonLINAC, PhotonSVDKernel
from ._base_pencilbeam import PencilBeamEngineAbstract


logger = logging.getLogger(__name__)


class DijSamplingConfig(TypedDict):
    """Properties for Dij sampling configuration."""

    rel_dose_threshold: float
    lat_cut_off: float
    type: Literal["radius", "depth"]
    delta_rad_depth: float


class PhotonPencilBeamSVDEngine(PencilBeamEngineAbstract):
    """
    Implementation of a pencil beam dose calculation engine for photons.

    The implementation is based on the Singular-value decomposition (SVD)
    method by Bortfeld, Schlegel & Rhein (1993).

    Parameters
    ----------
    pln : PhotonPlan
        A photon plan object.

    Attributes
    ----------
    use_custom_primary_photon_fluence : bool
        Use custom primary photon fluence.
    kernel_cutoff : float
        Kernel cutoff.
    random_seed : int
        Random seed.
    int_conv_resolution : float
        Intensity convolution resolution.
    enable_dij_sampling : bool
        Enable Dij sampling.
    dij_sampling : DijSamplingConfig
        Dij sampling configuration.
    """

    short_name = "SVDPB"
    name = "SVD Pencil Beam"
    possible_radiation_modes = ["photons"]

    use_custom_primary_photon_fluence: bool
    kernel_cutoff: float
    random_seed: int
    int_conv_resolution: float = 0.5
    enable_dij_sampling: bool = True
    dij_sampling: DijSamplingConfig

    def __init__(self, pln: PhotonPlan):
        self.use_custom_primary_photon_fluence = False
        self.kernel_cutoff = np.inf
        self.random_seed = 0
        self.int_conv_resolution = 0.5
        self.enable_dij_sampling = True
        self.dij_sampling = DijSamplingConfig(
            rel_dose_threshold=0.01, lat_cut_off=20, type="radius", delta_rad_depth=5
        )

        super().__init__(pln)

        # Protected/Private attributes (equivalent to SetAccess = protected)
        self._is_field_based_dose_calc = None  # will be set
        self._field_width = None  # will be obtained during calculation

        self._collimation = None  # collimation structure from DICOM import

    def _init_dose_calc(self, ct, cst, stf) -> dict[str, Any]:
        # TODO: What is "s" here?
        # self._is_field_based_dose_calc = any(str(s['bixelWidth']) == 'field' for s in stf)

        dij = super()._init_dose_calc(ct, cst, stf)

        # dij = []

        # checks of values
        # matrad here tests the kernel cutoff against the tabulated kernels
        # pyRadPlan, however, is more flexible and allows energy-specific kernels,
        # so we can't check this here but only when we load the energy-kernel?
        if self.kernel_cutoff < self.geometric_lateral_cutoff:
            logger.warning(
                "Kernel cutoff smaller than the geometric lateral cutoff. Using geometric cutoff."
            )

        # matRad does the gaussian filter here, but can we do that here?
        # We should be as flexible as to allow beams with different energies / penumbras / kernels
        # moved the kernel filtering to beam initialization

        # Initialize random number generator
        random.seed(self.random_seed)

        return dij

    def _init_beam(self, beam_info, ct, cst, stf, i):
        """
        Initialize a beam for pencil beam dose calculation.

        Parameters
        ----------
        beam_info : dict
            Beam information struct.
        ct : np.ndarray
            MatRad CT struct.
        cst : np.ndarray
            MatRad steering information struct.
        stf : np.ndarray
            MatRad steering information struct.
        i : int
            Index of beam.

        Returns
        -------
        dict
            Updated beam information struct.
        """

        beam_info = super()._init_beam(beam_info, ct, cst, stf, i)

        field_based_dose_calc = False
        if self._is_field_based_dose_calc:
            logger.debug("Enabling field-based dose calculation for beam %d!", i)
            self.int_conv_resolution = self._collimation["conv_resolution"]
            field_width = self._collimation["field_width"]
            field_based_dose_calc = True
        else:
            logger.debug("Enabling bixel-based dose calculation for beam %d!", i)
            field_width = beam_info["beam"]["bixel_width"]

        beam_info["field_based_dose_calc"] = field_based_dose_calc

        # TODO: obtain maximum field limits

        field_limit = np.ceil(field_width / (2 * self.int_conv_resolution))
        # TODO: should this be +1 as end or is it correct? In matRad it is equivalent to no +1
        field_grid = self.int_conv_resolution * np.arange(-field_limit, field_limit)
        beam_info["f_x"], beam_info["f_z"] = np.meshgrid(field_grid, field_grid, indexing="xy")

        # Get the kernel
        beamlets = [beamlet for ray in beam_info["beam"]["rays"] for beamlet in ray["beamlets"]]

        energies = np.unique([beamlet["energy"] for beamlet in beamlets])

        if len(energies) > 1:
            raise ValueError("Different energies in one photon beam not supported yet.")

        energy = energies[0]

        kernel = cast(PhotonLINAC, self._machine).get_kernel_by_energy(energy)

        if self.kernel_cutoff > kernel.kernel_pos[-1]:
            logger.info(
                "Kernel cutoff (%f mm) is larger than the beam's kernel range (%f mm)."
                " Using kernel range.",
                self.kernel_cutoff,
                kernel.kernel_pos[-1],
            )
            kernel_cutoff = kernel.kernel_pos[-1]
        else:
            kernel_cutoff = self.kernel_cutoff

        sigma_gauss = kernel.penumbra / np.sqrt(8 * np.log(2))  # [mm]

        # use 5 times sigma as the limits for the gaussian convolution
        gauss_limit = np.ceil(5 * sigma_gauss / self.int_conv_resolution)
        gauss_grid = self.int_conv_resolution * np.arange(-gauss_limit, gauss_limit)

        gauss_filter_x, gauss_filter_z = np.meshgrid(gauss_grid, gauss_grid, indexing="xy")
        gauss_filter = (
            1.0
            / (2 * np.pi * sigma_gauss**2 / self.int_conv_resolution**2)
            * np.exp(-(gauss_filter_x**2 + gauss_filter_z**2) / (2 * sigma_gauss**2))
        )

        gauss_conv_size = 2 * (field_limit + gauss_limit).astype(int)

        beam_info["gauss_conv_size"] = gauss_conv_size
        beam_info["gauss_filter"] = gauss_filter

        # get kernel size and distances
        kernel_limit = np.ceil(kernel_cutoff / self.int_conv_resolution)
        kernel_grid = self.int_conv_resolution * np.arange(-kernel_limit, kernel_limit)

        kernel_x, kernel_z = np.meshgrid(kernel_grid, kernel_grid, indexing="xy")

        # calculate also the total size and distance as we need this during convolution extensively
        kernel_conv_limit = field_limit + gauss_limit + kernel_limit
        kernel_conv_grid = self.int_conv_resolution * np.arange(
            -kernel_conv_limit, kernel_conv_limit
        )
        conv_mx_x, conv_mx_z = np.meshgrid(kernel_conv_grid, kernel_conv_grid, indexing="xy")

        kernel_conv_size = 2 * kernel_conv_limit.astype(int)

        effective_lateral_cut_off = self.geometric_lateral_cutoff + field_width / np.sqrt(2)
        beam_info["effective_lateral_cut_off"] = effective_lateral_cut_off

        if not field_based_dose_calc:
            n = np.floor(field_width / self.int_conv_resolution).astype(int)
            f_pre = np.ones((n, n), dtype=np.float32)

            if not self.use_custom_primary_photon_fluence:
                f_pre = fft.ifft2(
                    fft.fft2(f_pre, (gauss_conv_size, gauss_conv_size))
                    * fft.fft2(gauss_filter, (gauss_conv_size, gauss_conv_size))
                )
                f_pre = np.real(f_pre)

        # get index of central ray or closest to the central ray
        center = np.argmin(
            np.sum(
                np.array([ray["ray_pos_bev"] for ray in beam_info["beam"]["rays"]]) ** 2, axis=1
            )
        )

        center_ssd = beam_info["beam"]["rays"][center]["SSD"]

        # get correct kernel for given SSD at central ray
        kernels_at_ssd = kernel.get_kernels_at_ssd(center_ssd)

        # Display console message
        logger.info(
            "Kernel SSD = %g mm using %d components", center_ssd, kernel.num_kernel_components
        )

        # Get Interpolators
        # TODO: need scipy interpolate here probably
        kernel_mxs = np.apply_along_axis(
            lambda x: np.interp(
                np.sqrt(kernel_x**2 + kernel_z**2), kernel.kernel_pos, x, left=0.0, right=0.0
            ),
            axis=1,
            arr=kernels_at_ssd,
        )

        beam_info["kernel"] = kernel
        beam_info["kernel_mxs"] = kernel_mxs
        beam_info["f_pre"] = f_pre
        # beam_info["kernel_xz"] = (kernel_x.ravel(), kernel_z.ravel())
        # beam_info["conv_mx_xz"] = (conv_mx_x, conv_mx_z)
        beam_info["kernel_conv_grid"] = kernel_conv_grid
        beam_info["kernel_conv_size"] = kernel_conv_size

        kernel_interpolators = self._get_kernel_interpolators(beam_info, f_pre)
        beam_info["kernel_interpolators"] = kernel_interpolators

        return beam_info

    def _compute_bixel(self, curr_ray: dict[str], k: int) -> dict[str, Any]:
        """
        PyRadPlan photon dose calculation for a single bixel.

        call
            bixel = self.computeBixel(currRay,k)
        """
        bixel = {}

        kernel = cast(PhotonSVDKernel, curr_ray["kernel"])

        m = kernel.m
        betas = kernel.kernel_betas.reshape((-1, 1))
        rd = curr_ray["rad_depths"].reshape((1, -1))
        interpolators = cast(list[RegularGridInterpolator], curr_ray["kernel_interpolators"])
        iso_lat_dists = curr_ray["iso_lat_dists"]
        geo_depths = curr_ray["geo_depths"]
        sad = curr_ray["sad"]

        dose_component = betas / (betas - m) * (np.exp(-m * rd) - np.exp(-betas * rd))

        interpolated_kernels = [interp(iso_lat_dists) for interp in interpolators]

        for c, interp in enumerate(interpolated_kernels):
            dose_component[c, :] *= interp

        bixel_dose = np.sum(dose_component, axis=0)

        bixel_dose *= ((sad) / geo_depths) ** 2

        bixel["physical_dose"] = bixel_dose
        bixel["weight"] = curr_ray["beamlets"][k]["weight"]
        bixel["ix"] = curr_ray["ix"]

        return bixel

    def _get_kernel_interpolators(self, beam_info: dict[str], f: np.ndarray) -> list[Callable]:
        """Get kernel interpolator for photon dose calculation."""

        num_kernels = cast(PhotonSVDKernel, beam_info["kernel"]).num_kernel_components
        conv_size = beam_info["kernel_conv_size"]
        kernel_mxs = beam_info["kernel_mxs"]
        conv_grid = beam_info["kernel_conv_grid"]

        interpolators = [None] * num_kernels
        for c in range(num_kernels):
            conv_mx = np.real(
                fft.ifft2(
                    fft.fft2(f, (conv_size, conv_size))
                    * fft.fft2(kernel_mxs[c], (conv_size, conv_size))
                )
            )
            interpolators[c] = RegularGridInterpolator((conv_grid, conv_grid), conv_mx)

        return interpolators

    def _sample_dij(self, ix, bixel_dose, rad_depth_v, rad_distances_sq, bixel_width):
        """Perform lateral sampling of the beam."""

        raise NotImplementedError("This method is not implemented yet.")

    def _init_ray(self, beam_info: dict[str], j: int) -> dict[str]:
        """Initialize the current ray."""

        ray = super()._init_ray(beam_info, j)

        ray["kernel"] = beam_info["kernel"]

        if self.use_custom_primary_photon_fluence or beam_info["field_based_dose_calc"]:
            if beam_info["field_based_dose_calc"]:
                f = ray["shape"]
            else:
                f = beam_info["f_pre"]

            f = cast(np.ndarray, f)  # Typing

            primary_fluence = cast(PhotonSVDKernel, beam_info["kernel"]).primary_fluence
            r = np.sqrt(
                (beam_info["f_x"] - ray["ray_pos_bev"][0]) ** 2
                + (beam_info["f_x"] - ray["ray_pos_bev"][2]) ** 2
            )
            fx = f * np.interp(r, primary_fluence[:, 0], primary_fluence[:, 1])

            n = beam_info["gauss_conv_size"]
            gauss_filter = beam_info["gauss_filter"]

            fx = np.real(fft.ifft2(fft.fft2(fx, (n, n)) * fft.fft2(gauss_filter, (n, n))))

            ray["kernel_interpolators"] = self._get_kernel_interpolators(beam_info, fx)
        else:
            ray["kernel_interpolators"] = beam_info["kernel_interpolators"]

        return ray



"""Interface for voxel geometry ray tracers."""

from abc import ABC, abstractmethod
from typing import Union, Any
import logging
import time

import numpy as np
import SimpleITK as sitk

from pyRadPlan.core.np2sitk import linear_indices_to_image_coordinates
from pyRadPlan.geometry import lps
from pyRadPlan.stf._beam import Beam

from ._perf import fast_spatial_circle_lookup

logger = logging.getLogger(__name__)


class RayTracerBase(ABC):
    """Base class for all ray tracers."""

    lateral_cut_off: float
    precision: np.dtype

    @property
    def cubes(self):
        """CT or other abritrary cubes of similar resolution to be traced."""
        return self._cubes

    @cubes.setter
    def cubes(self, cubes: Union[sitk.Image, list[sitk.Image]]):
        if not isinstance(cubes, list):
            cubes = [cubes]
        self._cubes = cubes
        self._initialize_geometry()
        self._coords = None

    def __init__(self, cubes: Union[sitk.Image, list[sitk.Image]]):
        self.lateral_cut_off = 50.0
        self.precision = np.float32
        self.cubes = cubes
        self._coords = None

    def trace_rays(
        self,
        isocenter: Union[list, np.ndarray],
        source_points: Union[list, np.ndarray],
        target_points: Union[list, np.ndarray],
    ) -> tuple[np.ndarray, np.ndarray, list[np.ndarray], np.ndarray, np.ndarray]:
        """
        Trace multiple rays through a cube.

        Parameters
        ----------
        isocenter : Union[list, np.ndarray]
            Isocenter coordinates (1x3) array or list
        source_points : Union[list, np.ndarray]
            Source points coordinates. (nx3) array or list
        target_points : Union[list, np.ndarray]
            Target points coordinates. (nx3) array or list

        Returns
        -------
        alphas : ndarray
            Array of alpha values for each ray
        lengths : ndarray
            Array of lengths for each ray
        rho : list[ndarray]
            Array of rho values for each ray and each cube
        d12 : ndarray
            Array of full length of each ray
        ix : ndarray
            Linear indices (in numpy ordering) of the voxels intersected by each ray

        Notes
        -----
        The default implementation loops over the trace_ray function. The separte implementation is
        here to enable more performant implementations for specific ray tracers, e.g. through
        vectorization.
        """

        # Assuming size function equivalent is numpy's shape attribute.
        num_rays = target_points.shape[0]
        num_sources = source_points.shape[0]

        if num_sources not in (num_rays, 1):
            # MatRad_Config.instance() and dispError equivalent in Python needs handling.
            raise (
                f"Number of source points ({num_sources}) needs to be one "
                f"or equal to number of target points ({num_rays})!"
            )
        if num_sources == 1:
            source_points = np.tile(source_points, (num_rays, 1))
            num_sources = num_rays

        alphas, lengths, rho, d12, ix = [], [], [], [], []
        for r in range(num_rays):
            alpha, l_val, rho_val, d12_val, ix_val = self.trace_ray(
                isocenter, source_points[r, :], target_points[r, :]
            )
            alphas.append(alpha)
            lengths.append(l_val)
            rho.append(rho_val)
            d12.append(d12_val)
            ix.append(ix_val)

        # Padding with NaN values
        maxnumval = max(len(x) for x in ix)

        def nanpad(x):
            return np.pad(x, (0, maxnumval - len(x)), constant_values=np.nan)

        alphas = [nanpad(alpha) for alpha in alphas]
        lengths = [nanpad(l_val) for l_val in lengths]
        ix = [nanpad(ix_val) for ix_val in ix]

        for c in range(len(self.cubes)):
            rho[c] = [nanpad(rho_val) for rho_val in rho[c]]

        return np.array(alphas), np.array(lengths), rho, np.array(d12), np.array(ix)

    @abstractmethod
    def trace_ray(
        self,
        isocenter: Union[list, np.ndarray],
        source_points: Union[list, np.ndarray],
        target_points: Union[list, np.ndarray],
    ) -> tuple[np.ndarray, np.ndarray, list[np.ndarray], np.ndarray, np.ndarray]:
        """
        Trace a single ray through cubes.

        Abstract Method to be implemented in subclasses.
        """

    def trace_cubes(self, beam: Union[dict[str, Any], Beam]) -> list[sitk.Image]:
        """
        Automatically calculate depth by tracing rays through cubes.

        Set up ray matrix with appropriate spacing to trace through
        all cubes, resulting in a cumulative sum of values in every voxel
        relative to the source. Will calculate cumulative sum on all of
        the supplied images.
        """

        if not isinstance(beam, Beam):
            beam = Beam.model_validate(beam)

        t_trace_start = time.perf_counter()
        logger.debug("Computing coordinates...")
        if self._coords is None:
            # Obtain coordinates
            cube_ix = np.arange(self.cubes[0].GetNumberOfPixels(), dtype=np.int64)
            self._coords = linear_indices_to_image_coordinates(
                cube_ix, self.cubes[0], index_type="sitk", dtype=self.precision
            )

        # obtain rotation matrix
        rot_mat = lps.get_beam_rotation_matrix(beam.gantry_angle, beam.couch_angle)

        # rotate coordinates
        coords = (self._coords - beam.iso_center) @ rot_mat - beam.source_point_bev
        t_trace_end = time.perf_counter()
        logger.debug("took %s seconds!", t_trace_end - t_trace_start)

        # central_ray_vector = np.array(iso_center) - np.array(source_point).reshape
        logger.debug("Setting up Ray matrix...")
        t_trace_start = time.perf_counter()

        ray_spacing = np.min(self._resolution) / np.sqrt(2.0, dtype=self.precision)
        ray_matrix_bev_y = (
            np.max(coords[:, 1]) + np.max(self._resolution) + beam.source_point_bev[1]
        )
        ray_matrix_scale = 1 + ray_matrix_bev_y / beam.sad

        # num_candidate_rays = 2 * np.ceil(500.0 / ray_spacing).astype(np.int64) + 1

        spacing_range = ray_spacing * np.arange(
            np.floor(-500.0 / ray_spacing), np.ceil(500.0 / ray_spacing) + 1, dtype=self.precision
        )
        candidate_ray_coords_x, candidate_ray_coords_z = np.meshgrid(spacing_range, spacing_range)

        # If we have reference positions, we use them to restrict the raytracing region
        reference_positions_bev = ray_matrix_scale * np.array(
            [ray.ray_pos_bev for ray in beam.rays]
        )

        # use a precompiled numba function to speed up the spatial lookup
        candidate_ray_mx = fast_spatial_circle_lookup(
            candidate_ray_coords_x,
            candidate_ray_coords_z,
            reference_positions_bev,
            self.lateral_cut_off,
        )

        # candidate_ray_mx = np.full(candidate_ray_coords_x.shape, False, dtype=np.bool)

        # for i in range(reference_positions_bev.shape[0]):
        #     notix = (candidate_ray_coords_x - reference_positions_bev[i, 0]) ** 2 + (
        #         candidate_ray_coords_z - reference_positions_bev[i, 2]
        #     ) ** 2 <= self.lateral_cut_off**2
        #     candidate_ray_mx[notix] = True

        ray_matrix_bev = np.hstack(
            (
                candidate_ray_coords_x[candidate_ray_mx].reshape(-1, 1),
                ray_matrix_bev_y
                * np.ones(np.sum(candidate_ray_mx), dtype=self.precision).reshape(-1, 1),
                candidate_ray_coords_z[candidate_ray_mx].reshape(-1, 1),
            )
        )

        ray_matrix_lps = ray_matrix_bev @ rot_mat.T

        t_trace_end = time.perf_counter()
        logger.debug("took %s seconds!", t_trace_end - t_trace_start)

        logger.debug("Tracing %d rays through the cubes", np.count_nonzero(candidate_ray_mx))

        t_trace_start = time.perf_counter()
        _, lengths, rho, d12, ix = self.trace_rays(
            beam.iso_center.reshape(1, 3), beam.source_point.reshape(1, 3), ray_matrix_lps
        )
        t_trace_end = time.perf_counter()

        logger.debug("Cube ray tracing took %s seconds...", t_trace_end - t_trace_start)

        # Now we compute which rays will respectively give the voxel value for radiological depth
        valid_ix = np.isfinite(ix)

        scale_factor = np.zeros_like(ix, dtype=self.precision)
        scale_factor[valid_ix] = (ray_matrix_bev_y + beam.sad) / coords[ix[valid_ix], 1]

        x_dist = np.full_like(ix, np.nan, dtype=self.precision)
        z_dist = np.full_like(ix, np.nan, dtype=self.precision)

        x_dist[valid_ix] = coords[ix[valid_ix], 0] * scale_factor[valid_ix]
        x_dist = x_dist - ray_matrix_bev[:, 0, np.newaxis]

        z_dist[valid_ix] = coords[ix[valid_ix], 2] * scale_factor[valid_ix]
        z_dist = z_dist - ray_matrix_bev[:, 2, np.newaxis]

        ray_selection = ray_spacing / 2.0

        ix_remember_from_tracing = (
            (x_dist > -ray_selection)
            & (x_dist <= ray_selection)
            & (z_dist > -ray_selection)
            & (z_dist <= ray_selection)
        )
        t_remember_end = time.perf_counter()

        logger.debug(
            "Found %d ray indices for radiological depth calculation (took %s seconds)",
            np.count_nonzero(ix_remember_from_tracing),
            t_remember_end - t_trace_end,
        )
        rad_depth_cubes = [
            np.nan * np.ones_like(sitk.GetArrayViewFromImage(cube), dtype=self.precision)
            for cube in self.cubes
        ]

        for i, cube in enumerate(rad_depth_cubes):
            rel_depths = lengths * rho[i]
            rel_depths = np.cumsum(rel_depths, axis=1) - rel_depths / 2.0

            try:
                ix_assign = np.unravel_index(ix[ix_remember_from_tracing], cube.shape, order="F")
            except (ValueError, IndexError):
                logger.error(
                    "Error in unraveling indices from raytracing. Trying to recover...",
                    exc_info=True,
                )
                tmp_ix = ix[ix_remember_from_tracing]
                rel_depths = rel_depths[ix_remember_from_tracing]

                wrong_values = np.logical_or(tmp_ix < 0, tmp_ix >= cube.size)
                tmp_ix = tmp_ix[~wrong_values]
                rel_depths = rel_depths[~wrong_values]

                # Remove the wrong values
                ix_assign = np.unravel_index(tmp_ix, cube.shape, order="F")
                logger.info(
                    "Recovered %d indices for radiological depth cube",
                    np.count_nonzero(wrong_values),
                )
                cube[ix_assign] = rel_depths
            else:
                cube[ix_assign] = rel_depths[ix_remember_from_tracing]

            rad_depth_cubes[i] = sitk.GetImageFromArray(cube)
            rad_depth_cubes[i].CopyInformation(self.cubes[i])

        t_createcubes_end = time.perf_counter()

        logger.debug(
            "Radiological depth cube filling took %s seconds",
            t_createcubes_end - t_remember_end,
        )

        return rad_depth_cubes
        # scale_factor[valid_ix] = lengths[valid_ix] / d12[valid_ix]

    @abstractmethod
    def _initialize_geometry(self):
        """
        Initialize geometry of the ray tracer.

        Will be automatically called when the cubes are set.
        """



from ._voi import VOI, OAR, Target, HelperVOI, ExternalVOI, create_voi, validate_voi
from ._cst import StructureSet, create_cst, validate_cst

__all__ = [
    "StructureSet",
    "create_cst",
    "validate_cst",
    "VOI",
    "OAR",
    "Target",
    "HelperVOI",
    "ExternalVOI",
    "create_voi",
    "validate_voi",
]



"""
Nominal Scenario Model.

This module contains the NominalScenario class, which represents an example scenario.

Example
-------
    $ python -m pyRadPlan.scenarios._nominal

    # Usage Example
    # scenario = NominalScenario()
    # scenario.list_all_scenarios()
"""

from typing import ClassVar

import warnings
import numpy as np
from pyRadPlan.scenarios import ScenarioModel


# Example subclass
class NominalScenario(ScenarioModel):
    """
    NominalScenario class represents an example scenario.

    Attributes
    ----------
    name : str
        The name of the scenario.
    short_name : str
        The short name of the scenario.

    Methods
    -------
    update_scenarios()
        Updates the scenarios.
    extract_single_scenario(scen_num: int) -> ScenarioModel
        Extracts a single scenario.
    """

    name: ClassVar[str] = "Nominal Scenario"
    short_name: ClassVar[str] = "nomScen"

    @property
    def ndim(self) -> int:
        """Dimensionality of the scenario model."""
        return 1

    def update_scenarios(self) -> np.ndarray[float]:
        num_of_ct_scen = len(self.ct_scen_prob)
        self._num_of_available_ct_scen = num_of_ct_scen

        # Scenario weight
        self._scen_weight = np.ones(num_of_ct_scen).astype(float) / float(num_of_ct_scen)
        self._scen_weight = [prob[1] for prob in self.ct_scen_prob]
        self._ct_scen_ix = [prob[0] for prob in self.ct_scen_prob]

        # set variables
        self._tot_num_shift_scen = 1
        self._tot_num_range_scen = 1
        self._tot_num_scen = num_of_ct_scen

        # Individual shifts
        self._rel_range_shift = np.zeros(num_of_ct_scen)
        self._abs_range_shift = np.zeros(num_of_ct_scen)
        self._iso_shift = np.zeros((num_of_ct_scen, 3))

        # Probability matrices
        self._scen_for_prob = np.hstack(
            (np.array(self.ct_scen_prob)[:, 0].reshape(-1, 1), np.zeros((num_of_ct_scen, 5)))
        )  # Realization matrix
        self._scen_prob = np.array(self.ct_scen_prob)[:, 1]  # Probabilities for each scenario

        # Mask for scenario selection
        self._scen_mask = np.zeros(
            (self._num_of_available_ct_scen, self._tot_num_shift_scen, self._tot_num_range_scen),
            dtype=bool,
        )
        self._scen_mask[self._ct_scen_ix, :, :] = True

        # generic code
        x = np.where(self._scen_mask)
        self._linear_mask = np.column_stack(x)
        tot_num_scen = np.sum(self._scen_mask)

        # Get Scenario probability
        covariance = np.diag(
            np.hstack((np.array(self.shift_sd), self.range_abs_sd, self.range_rel_sd / 100.0)) ** 2
        )
        d = float(covariance.shape[0])
        cs = np.linalg.cholesky(covariance)
        tmp_standardized = np.linalg.solve(cs, self._scen_for_prob[:, 1:].T).T
        tmp_scen_prob = (
            (2 * np.pi) ** (-d / 2)
            * np.exp(-0.5 * np.sum(tmp_standardized**2, axis=1))
            / np.prod(np.diag(cs))
        )

        # Multiply with 4D phase probability
        self._scen_prob = self._scen_weight * tmp_scen_prob

        # Get relative (normalized) weight of the scenario
        self._scen_weight = self._scen_prob / np.sum(self._scen_prob)

        # Return variable
        scenarios = self._scen_for_prob

        if int(tot_num_scen) != self._tot_num_scen:
            warnings.warn(
                (
                    "Check Implementation of Total Scenario computation - "
                    f"given {self._tot_num_scen} but found {tot_num_scen}!"
                )
            )
            self._tot_num_scen = tot_num_scen

        return scenarios

    # TODO: Finalize implementation to extract according to ct number
    def extract_single_scenario(self, scen_num: int) -> "ScenarioModel":
        # Example logic
        if scen_num != 0:
            raise NotImplementedError(
                "Currently the Nominal Scenario Model is only implemented for single ct scenarios!"
            )
        return self



"""Helpers for conversion between numpy and SimpleITK."""

from typing import Literal
import SimpleITK as sitk
import numpy as np
from ._grids import Grid


def sitk_mask_to_linear_indices(mask: sitk.Image, order="sitk") -> np.ndarray:
    """
    Convert a SimpleITK mask to linear indices.

    Parameters
    ----------
    mask : sitk.Image
        The SimpleITK image mask to be converted.
    order : str, optional
        The ordering of the indices. Can be 'sitk' for Fortran-like index
        ordering or 'numpy' for C-like index ordering. Default is 'sitk'.

    Returns
    -------
    np.ndarray
        A 1D numpy array of linear indices where the mask is non-zero.

    Raises
    ------
    ValueError
        If the ordering is not 'sitk' or 'numpy'.
    """

    arr = sitk.GetArrayViewFromImage(mask)
    if order == "sitk":
        return np.nonzero(arr.ravel(order="F"))[0]
    if order == "numpy":
        return np.flatnonzero(arr)
    raise ValueError("Invalid ordering. Must be 'sitk' or 'numpy'.")


def linear_indices_to_sitk_mask(
    indices: np.ndarray, ref_image: sitk.Image, order="sitk"
) -> sitk.Image:
    """
    Convert linear indices to a SimpleITK mask.

    Parameters
    ----------
    indices : np.ndarray
        A 1D numpy array of linear indices where the mask is non-zero.
    ref_image : sitk.Image
        The reference image on which the mask is defined.
    order : str, optional
        The ordering of the indices. Can be 'sitk' for Fortran-like index
        ordering or 'numpy' for C-like index ordering. Default is 'sitk'.

    Returns
    -------
    sitk.Image
        A SimpleITK image mask with the indices set to non-zero.

    Raises
    ------
    ValueError
        If the ordering is not 'sitk' or 'numpy'.
    """

    arr = np.zeros_like(sitk.GetArrayViewFromImage(ref_image), dtype=np.uint8)

    if order == "sitk":
        arr.T.flat[indices] = 1
    elif order == "numpy":
        arr.flat[indices] = 1
    else:
        raise ValueError("Invalid ordering. Must be 'sitk' or 'numpy'.")

    mask = sitk.GetImageFromArray(arr)
    mask.CopyInformation(ref_image)

    return mask


def linear_indices_to_grid_coordinates(
    indices: np.ndarray,
    grid: Grid,
    index_type: Literal["numpy", "sitk"] = "numpy",
    dtype: np.dtype = np.float64,
) -> np.ndarray:
    """
    Convert linear indices to gridcoordinates.

    Parameters
    ----------
    indices : np.ndarray
        A 1D numpy array of linear indices where the mask is non-zero.
    grid : Grid
        The image grid on which the indices lie.
    index_type : Literal["numpy", "sitk"], optional
        The ordering of the indices. Can be 'sitk' for Fortran-like index
        ordering or 'numpy' for C-like index ordering. Default is 'numpy'.
    dtype : np.dtype, optional
        The data type of the output coordinates. Default is np.float64.

    Returns
    -------
    np.ndarray
        A 2D numpy array of image coordinates.
    """

    # this is a manual reimplementation of np.unravel_index
    # to avoid the overhead of creating a tuple of arrays
    if index_type == "numpy":
        _, d1, d2 = grid.dimensions[::-1]
        order = [0, 1, 2]
    elif index_type == "sitk":
        _, d1, d2 = grid.dimensions
        order = [2, 1, 0]
    else:
        raise ValueError("Invalid index type. Must be 'numpy' or 'sitk'.")

    v = np.empty((3, np.asarray(indices).size), dtype=dtype)
    tmp, v[order[0]] = np.divmod(indices, d2)
    v[order[2]], v[order[1]] = np.divmod(tmp, d1)

    spacing_diag = np.diag(
        [grid.resolution["x"], grid.resolution["y"], grid.resolution["z"]]
    ).astype(dtype)
    origin = grid.origin.astype(dtype)

    physical_point = origin + np.matmul(np.matmul(grid.direction, spacing_diag), v).T

    return physical_point


def linear_indices_to_image_coordinates(
    indices: np.ndarray,
    image: sitk.Image,
    index_type: Literal["numpy", "sitk"] = "numpy",
    dtype: np.dtype = np.float64,
) -> np.ndarray:
    """
    Convert linear indices to image coordinates.

    Parameters
    ----------
    indices : np.ndarray
        A 1D numpy array of linear indices where the mask is non-zero.
    image : sitk.Image
        The reference image on which the mask is defined.
    index_type : Literal["numpy", "sitk"], optional
        The ordering of the indices. Can be 'sitk' for Fortran-like index
        ordering or 'numpy' for C-like index ordering. Default is 'numpy'.
    dtype : np.dtype, optional
        The data type of the output coordinates. Default is np.float64.

    Returns
    -------
    np.ndarray
        A 2D numpy array of image coordinates.
    """

    grid = Grid.from_sitk_image(image)
    return linear_indices_to_grid_coordinates(indices, grid, index_type, dtype=dtype)



import pytest
import sys

if sys.version_info < (3, 10):
    import importlib_resources as resources  # Backport for older versions
else:
    from importlib import resources  # Standard from Python 3.9+

import pymatreader
import numpy as np

from pyRadPlan.ct import validate_ct
from pyRadPlan.cst import validate_cst
from pyRadPlan.plan import IonPlan
from pyRadPlan.stf import SteeringInformation, create_stf
from pyRadPlan.stf.generators import StfGeneratorIMPT


@pytest.fixture
def sample_stf_snake():
    beam = {
        "gantry_angle": 0,
        "couch_angle": 0,
        "bixel_width": 1,
        "radiation_mode": "protons",
        "machine": "Generic",
        "sad": 100000,
        "iso_center": [0, 0, 0],
        "rays": [
            {"ray_pos_bev": ([0, 1, 2]), "ray_pos": ([5, 6, 7])},
            {"ray_pos_bev": ([0, 1, 2]), "ray_pos": ([5, 6, 7])},
        ],
        "source_point_bev": [0, -10000, 0],
        "source_point": [0, 0, 0],
        "longitudinal_spot_spacing": 1,
    }
    return beam


@pytest.fixture
def sample_stf_camel():
    beam = {
        "gantryAngle": 0,
        "couchAngle": 0,
        "bixelWidth": 1,
        "radiationMode": "protons",
        "machine": "Generic",
        "sad": 100000,
        "isoCenter": [0, 0, 0],
        "rays": [
            {"rayPos_bev": ([0, 1, 2]), "rayPos": ([5, 6, 7])},
            {"rayPos_bev": ([0, 1, 2]), "rayPos": ([5, 6, 7])},
        ],
        "sourcePointBev": [0, -10000, 0],
        "sourcePoint": [0, 0, 0],
        "longitudinalSpotSpacing": 1,
    }
    return beam


def test_create_stf_from_snake_case(sample_stf_snake):
    beam1 = sample_stf_snake
    input1 = [beam1, beam1]
    input2 = {"beams": [beam1, beam1]}

    # Testing for list[Dict]
    stf1 = create_stf(input1)
    assert isinstance(stf1, SteeringInformation)
    assert stf1.bixel_beam_index_map.size == stf1.total_number_of_bixels
    assert stf1.bixel_ray_index_per_beam_map.size == stf1.total_number_of_bixels
    assert stf1.bixel_index_per_beam_map.size == stf1.total_number_of_bixels

    tot_num_bixels = 0
    for i in range(len(input1)):
        assert hasattr(stf1.beams[i], "couch_angle")
        assert hasattr(stf1.beams[i].rays[0], "ray_pos_bev")
        assert all(hasattr(ray, "ray_pos_bev") for ray in stf1.beams[i].rays)
        assert all(hasattr(ray, "beamlets") for ray in stf1.beams[i].rays)
        assert all(isinstance(ray.beamlets, list) for ray in stf1.beams[i].rays)

        assert stf1.beams[i].bixel_ray_map.size == stf1.beams[i].total_number_of_bixels
        assert np.array_equal(
            stf1.beams[i].bixel_ray_map,
            stf1.bixel_ray_index_per_beam_map[
                tot_num_bixels : tot_num_bixels + stf1.beams[i].total_number_of_bixels
            ],
        )
        tot_num_bixels += stf1.beams[i].total_number_of_bixels

    # Testing for Dict[list[Dict]]
    stf2 = create_stf(input2)
    for i in range(len(input1)):
        assert hasattr(stf2.beams[i], "couch_angle")
        assert hasattr(stf2.beams[i].rays[0], "ray_pos_bev")
        assert all(hasattr(ray, "ray_pos_bev") for ray in stf2.beams[i].rays)
        assert all(hasattr(ray, "beamlets") for ray in stf2.beams[i].rays)
        assert all(isinstance(ray.beamlets, list) for ray in stf2.beams[i].rays)


def test_create_stf_from_camel_case(sample_stf_camel):
    beam1 = sample_stf_camel
    input1 = [beam1, beam1]
    input2 = {"beams": [beam1, beam1]}

    # Testing for list[Dict]
    stf1 = create_stf(input1)
    assert isinstance(stf1, SteeringInformation)
    for i in range(len(input1)):
        assert hasattr(stf1.beams[i], "couch_angle")
        assert hasattr(stf1.beams[i].rays[0], "ray_pos_bev")
        assert all(hasattr(ray, "ray_pos_bev") for ray in stf1.beams[i].rays)
        assert all(hasattr(ray, "beamlets") for ray in stf1.beams[i].rays)
        assert all(isinstance(ray.beamlets, list) for ray in stf1.beams[i].rays)

    # Testing for Dict[list[Dict]]
    stf2 = create_stf(input2)
    for i in range(len(input1)):
        assert hasattr(stf2.beams[i], "couch_angle")
        assert hasattr(stf2.beams[i].rays[0], "ray_pos_bev")
        assert all(hasattr(ray, "ray_pos_bev") for ray in stf2.beams[i].rays)
        assert all(hasattr(ray, "beamlets") for ray in stf2.beams[i].rays)
        assert all(isinstance(ray.beamlets, list) for ray in stf2.beams[i].rays)


def test_create_stf_from_stf(sample_stf_snake):
    beam1 = sample_stf_snake
    beam2 = beam1.copy()
    beam2["gantry_angle"] = 90
    input1 = [beam1, beam2]

    # Creating stf
    stf = SteeringInformation(beams=input1)
    for i, beam in enumerate(stf.beams):
        assert hasattr(beam, "couch_angle")
        assert hasattr(beam.rays[0], "ray_pos_bev")
        assert all(hasattr(ray, "ray_pos_bev") for ray in beam.rays)
        assert all(hasattr(ray, "beamlets") for ray in beam.rays)
        assert all(isinstance(ray.beamlets, list) for ray in beam.rays)

    # Creating stf from existing stf by SteeringInformation()
    stf = create_stf(stf)
    for i, beam in enumerate(stf.beams):
        assert hasattr(beam, "couch_angle")
        assert hasattr(beam.rays[0], "ray_pos_bev")
        assert all(hasattr(ray, "ray_pos_bev") for ray in beam.rays)
        assert all(hasattr(ray, "beamlets") for ray in beam.rays)
        assert all(isinstance(ray.beamlets, list) for ray in beam.rays)

    stf = create_stf(input1)
    assert isinstance(stf, SteeringInformation)
    for i, beam in enumerate(stf.beams):
        assert hasattr(beam, "couch_angle")
        assert hasattr(beam.rays[0], "ray_pos_bev")
        assert all(hasattr(ray, "ray_pos_bev") for ray in beam.rays)
        assert all(hasattr(ray, "beamlets") for ray in beam.rays)
        assert all(isinstance(ray.beamlets, list) for ray in beam.rays)


def test_stfgen_to_stf():
    files = resources.files("pyRadPlan.data.phantoms")
    path = files.joinpath("TG119.mat")
    tmp = pymatreader.read_mat(path)

    ct = validate_ct(tmp["ct"])
    cst = validate_cst(tmp["cst"], ct=ct)
    pln = IonPlan(radiation_mode="protons", machine="Generic")

    stfgen = StfGeneratorIMPT(pln)
    stfgen.bixel_width = 5.0
    stfgen.gantry_angles = [0.0]

    stf_native = stfgen.generate(ct, cst)
    stf = create_stf(stf_native)

    assert stf is not None
    assert isinstance(stf, SteeringInformation)

    for i, beam in enumerate(stf.beams):
        assert hasattr(beam, "couch_angle")
        assert hasattr(beam.rays[0], "ray_pos_bev")
        assert all(hasattr(ray, "ray_pos_bev") for ray in beam.rays)
        assert all(hasattr(ray, "beamlets") for ray in beam.rays)
        assert all(isinstance(ray.beamlets, list) for ray in beam.rays)


def test_from_and_to_matrad():
    # from scipy.io import savemat

    files = resources.files("pyRadPlan.data.stf")
    path = files.joinpath("matRad_stf.mat")
    path_export = files.joinpath("export_stf.mat")
    matRad_file = pymatreader.read_mat(path)
    stf_from_matRad = create_stf(matRad_file)
    stf_to_matrad = stf_from_matRad.to_matrad()

    assert isinstance(stf_to_matrad, np.recarray)
    # savemat(path_export, stf_to_matrad, format='5', oned_as='column')


def test_from_matlab_one_beam():
    # Special case of one beam since pymatreader handles 1D arrays differently
    files = resources.files("pyRadPlan.data.stf")
    path = files.joinpath("matRad_stf_onebeam.mat")
    matRad_file = pymatreader.read_mat(path)
    stf_from_matRad = create_stf(matRad_file)
    assert stf_from_matRad.beams[0].gantry_angle == 0
    assert stf_from_matRad is not None

    assert stf_from_matRad.bixel_beam_index_map.size == stf_from_matRad.total_number_of_bixels
    assert (
        stf_from_matRad.bixel_ray_index_per_beam_map.size == stf_from_matRad.total_number_of_bixels
    )
    assert stf_from_matRad.bixel_index_per_beam_map.size == stf_from_matRad.total_number_of_bixels


def test_from_matlab_multi_beam():
    files = resources.files("pyRadPlan.data.stf")
    path = files.joinpath("matRad_stf.mat")
    matRad_file = pymatreader.read_mat(path)
    stf_from_matRad = create_stf(matRad_file)
    assert stf_from_matRad.beams[0].gantry_angle == 0
    assert stf_from_matRad is not None

    assert stf_from_matRad.bixel_beam_index_map.size == stf_from_matRad.total_number_of_bixels
    assert (
        stf_from_matRad.bixel_ray_index_per_beam_map.size == stf_from_matRad.total_number_of_bixels
    )
    assert stf_from_matRad.bixel_index_per_beam_map.size == stf_from_matRad.total_number_of_bixels

    tot_num_bixels = 0
    for i in range(stf_from_matRad.num_of_beams):
        assert (
            stf_from_matRad.beams[i].bixel_ray_map.size
            == stf_from_matRad.beams[i].total_number_of_bixels
        )
        assert np.array_equal(
            stf_from_matRad.beams[i].bixel_ray_map,
            stf_from_matRad.bixel_ray_index_per_beam_map[
                tot_num_bixels : tot_num_bixels + stf_from_matRad.beams[i].total_number_of_bixels
            ],
        )
        assert np.all(
            stf_from_matRad.bixel_beam_index_map[
                tot_num_bixels : tot_num_bixels + stf_from_matRad.beams[i].total_number_of_bixels
            ]
            == i
        )
        assert np.all(
            stf_from_matRad.bixel_index_per_beam_map[
                tot_num_bixels : tot_num_bixels + stf_from_matRad.beams[i].total_number_of_bixels
            ]
            == np.arange(stf_from_matRad.beams[i].total_number_of_bixels)
        )
        assert np.all(
            stf_from_matRad.beams[i].bixel_ray_map
            == stf_from_matRad.bixel_ray_index_per_beam_map[
                tot_num_bixels : tot_num_bixels + stf_from_matRad.beams[i].total_number_of_bixels
            ]
        )

        tot_num_bixels += stf_from_matRad.beams[i].total_number_of_bixels


# def test_create_stf_kwargs_snake():
#     stf1 = create_stf(Beam={"Beam":{"gantry_angle":0}})
#     assert isinstance(stf1, SteeringInformation)
#     assert stf1.gantry_angle == 0



"""
Contains the definition of the Plan class and its derived classes.

Available spezialized Plan classes are PhotonPlan and IonPlan.
"""

from abc import ABC
from typing import Dict, Any, List, Union, ClassVar
from copy import deepcopy

from pydantic import (
    Field,
    field_validator,
    ValidationError,
)
from pydantic.alias_generators import to_snake
from pyRadPlan.core import PyRadPlanBaseModel
from pyRadPlan.scenarios import ScenarioModel, create_scenario_model, validate_scenario_model


class Plan(PyRadPlanBaseModel, ABC):
    """
    Abstract base class for a treatment plan using PyRadPlanBaseModel.

    Attributes
    ----------
    prop_stf : Dict[str, Any]
        Properties of the stf.
    prop_opt : Dict[str, Any]
        Properties of the optimization.
    prop_dose_calc : Dict[str, Any]
        Properties of the dose calculation.
    prop_seq : Dict[str, Any]
        Properties for the sequencer
    num_of_fractions : int
        Number of fractions in the plan.
    machine : str
        Machine used for the plan.
    prescribed_dose : float
        Prescribed dose for the plan. Serves mainly as normalization value.
    radiation_mode : str
        Will return the radiation modality (e.g. photons or protons).
    """

    prop_stf: Dict[str, Any] = Field(default_factory=dict)
    prop_opt: Dict[str, Any] = Field(default_factory=dict)
    prop_dose_calc: Dict[str, Any] = Field(default_factory=dict)
    prop_seq: Dict[str, Any] = Field(default_factory=dict)
    num_of_fractions: int = Field(default=30, gt=0)
    machine: Union[Dict, str] = Field(default="Generic")
    prescribed_dose: float = Field(default=60.0, gt=0.0)
    mult_scen: ScenarioModel = Field(default_factory=create_scenario_model)

    # Abstract property handled by below validator
    radiation_mode: str

    @field_validator("radiation_mode", mode="after")
    @classmethod
    def validate_radiation_mode(cls, v: str) -> str:
        """
        Validate the radiation mode.

        Parameters
        ----------
        v : str
            The radiation mode value to be validated.

        Raises
        ------
        NotImplementedError
            This method should be overridden in derived classes.
        """
        raise NotImplementedError("This method should be overridden in derived classes")

    @field_validator("mult_scen", mode="before")
    @classmethod
    def _validate_mult_scen(
        cls, v: Union[Union[Dict[str, Any], ScenarioModel], str]
    ) -> ScenarioModel:
        """
        Validate the mult_scen attribute.

        Parameters
        ----------
        v : Union[Dict[str, Any], ScenarioModel]
            The mult_scen attribute to be validated.

        Returns
        -------
        ScenarioModel
            The validated mult_scen attribute.

        Raises
        ------
        ValueError
            If the mult_scen attribute is not a valid ScenarioModel object.
        """

        try:
            return validate_scenario_model(v)
        except ValueError as exc:
            raise ValidationError(
                "mult_scen must be a ScenarioModel object or respective dictionary"
            ) from exc

    @field_validator("prop_stf", "prop_opt", "prop_dose_calc", "prop_opt", mode="after")
    @classmethod
    def validate_prop(cls, v: Dict[str, Any]) -> Dict[str, Any]:
        """
        Validate the workflow property dictionaries.

        Will try to convert to snake_case if camelCase is used.

        Parameters
        ----------
        v : Dict[str, Any]
            The properties of the plan to be validated.

        Returns
        -------
        Dict[str, Any]
            The validated properties of the plan.
        """

        if not v:
            return {}

        # Convert camelCase to snake_case
        return {to_snake(k): v for k, v in v.items()}

    def to_matrad(self, context: str = "mat-file") -> Any:
        """
        Create a dictionary ready to save the Plan model to a mat-file.

        Returns
        -------
            Dict: A dictionary containing the data of the Plan model in a format suitable for
            saving to a mat-file.
        """

        pln_dict = super().to_matrad(context=context)
        pln_dict["numOfFractions"] = float(pln_dict["numOfFractions"])
        return pln_dict


class PhotonPlan(Plan):
    """
    Class for a photon treatment plan.

    Attributes
    ----------
    Inherits all attributes from Plan.

    Methods
    -------
    radiation_mode : str
        Returns the radiation mode as 'photons'.
    """

    radiation_mode: str = "photons"

    @field_validator("radiation_mode", mode="after")
    @classmethod
    def validate_radiation_mode(cls, v: str) -> str:
        """
        Validate the radiation mode for a PhotonPlan.

        Parameters
        ----------
        v : str
            The radiation mode to be validated.

        Returns
        -------
        str
            The validated radiation mode.

        Raises
        ------
        ValueError
            If the radiation mode is not "photons".
        """
        if v != "photons":
            raise ValueError('radiation_mode for PhotonPlan must be "photons"')
        return v


class IonPlan(Plan):
    """
    Class for an ion treatment plan.

    Attributes
    ----------
    ionType : str
        Type of ion used in the plan.
    Inherits all other attributes from Plan.

    Methods
    -------
    radiation_mode : str
        Returns the radiation mode as the ion type.
    """

    available_radiation_modes: ClassVar[List[str]] = ["protons", "helium", "carbon", "oxygen"]

    radiation_mode: str = Field(
        default="protons", pattern="^(protons|helium|carbon|oxygen)$", validate_default=True
    )

    @field_validator("radiation_mode", mode="after")
    @classmethod
    def validate_radiation_mode(cls, v: str) -> str:
        """
        Validate the radiation mode for IonPlan.

        Parameters
        ----------
        cls : class
            The class object.
        v : str
            The radiation mode to be validated.

        Returns
        -------
        str
            The validated radiation mode.

        Raises
        ------
        ValueError
            If the radiation mode is not one of the available radiation modes.
        """
        if v not in cls.available_radiation_modes:
            raise ValueError(
                f"radiation_mode for IonPlan must be one of {cls.available_radiation_modes}"
            )
        return v


def create_pln(data: Union[Dict[str, Any], Plan, None] = None, **kwargs) -> Plan:
    """
    Create a Plan object (factory function).

    Parameters
    ----------
    data : Union[Dict[str, Any], None]
        Dictionary containing the data to create the Plan object.
    **kwargs
        Arbitrary keyword arguments.

    Returns
    -------
    Plan
        A Plan object.

    Raises
    ------
    ValueError
        If the radiation mode is unknown or empty.
    """
    data = deepcopy(data)
    if data:
        # If data is already a Plan object, return it directly
        if isinstance(data, Plan):
            return data

        # obtain the radiation mode if we have a dictionary at our hands
        radiation_mode = data.get("radiation_mode")

        # Since we also allow camelCase, try to get radiationMode if radiation_mode is not set
        if radiation_mode is None:
            radiation_mode = data.get("radiationMode")

        if radiation_mode == "photons":
            return PhotonPlan.model_validate(data)
        # radiation_mode in ['protons', 'helium', 'carbon', 'oxygen']:
        return IonPlan.model_validate(data)
        # raise ValueError(f"Unknown radiation mode: {radiation_mode}")
    radiation_mode = kwargs.get("radiation_mode", "")
    if radiation_mode == "photons":
        return PhotonPlan(**kwargs)
    if radiation_mode in ["protons", "helium", "carbon", "oxygen"]:
        return IonPlan(**kwargs)
    raise ValueError(f"Unknown radiation mode: {radiation_mode}")


def validate_pln(plan: Union[Dict[str, Any], Plan, None] = None, **kwargs) -> Plan:
    """
    Validate and create a Plan object.

    Synonym to create_pln but should be used in validation context.

    Parameters
    ----------
    plan : Union[Dict[str, Any], Plan, None], optional
        Dictionary containing the data to create the Plan object, by default None.
    **kwargs
        Arbitrary keyword arguments.

    Returns
    -------
    Plan
        A validated Plan object.

    Raises
    ------
    ValueError
        If the radiation mode is unknown or empty.
    """
    return create_pln(plan, **kwargs)


if __name__ == "__main__":
    scen = create_scenario_model("nomScen")
    scen_dict_camel = scen.to_matrad()
    scen_dict_snake = scen.to_dict()

    pln_dict_camel = {
        "radiationMode": "photons",  # either photons / protons / carbon
        "machine": "Generic",
        "numOfFractions": 30,
        "prescribedDose": 60.0,
        "propStf": {},
        # dose calculation settings
        "propDoseCalc": {},
        # optimization settings
        "propOpt": {},
        "propSeq": {},
        "multScen": scen_dict_camel,
    }

    pln = create_pln(pln_dict_camel)



"""Module providing the geometry / stf generators."""

from ._base import StfGeneratorBase
from ._externalbeam import StfGeneratorExternalBeamRayBixel, StfGeneratorExternalBeam
from ._ions import StfGeneratorIonSingleSpot, StfGeneratorIMPT
from ._photons import StfGeneratorPhotonIMRT, StfGeneratorPhotonCollimatedSquareFields

from ._factory import get_generator, get_available_generators, register_generator

register_generator(StfGeneratorIMPT)
register_generator(StfGeneratorPhotonIMRT)
register_generator(StfGeneratorIonSingleSpot)
register_generator(StfGeneratorPhotonCollimatedSquareFields)

__all__ = [
    "StfGeneratorBase",
    "StfGeneratorExternalBeam",
    "StfGeneratorExternalBeamRayBixel",
    "StfGeneratorPhotonIMRT",
    "StfGeneratorPhotonCollimatedSquareFields",
    "StfGeneratorIonSingleSpot",
    "StfGeneratorIMPT",
    "get_generator",
    "get_available_generators",
    "register_generator",
]



import numpy as np
from pyRadPlan.ct import default_hlut


def test_default_hlut():
    """Test default_hlut."""
    hlut = default_hlut()
    assert isinstance(hlut, np.ndarray)
    assert hlut.shape[1] == 2
    assert hlut.shape[0] >= 1



"""Geometry module."""

from .lps import get_beam_rotation_matrix, get_couch_rotation_matrix, get_gantry_rotation_matrix

__all__ = [
    "get_beam_rotation_matrix",
    "get_couch_rotation_matrix",
    "get_gantry_rotation_matrix",
]



"""Solver Base Classes for Planning Problems."""

from typing import ClassVar, Callable
from abc import ABC, abstractmethod

import numpy as np
from numpy.typing import ArrayLike


class SolverBase(ABC):
    """
    Abstract Base Class for Solver Implementations / Interfaces.

    Attributes
    ----------
    name : ClassVar[str]
        Full name of the solver
    short_name : ClassVar[str]
        Short name of the solver
    max_time : float, default=3600
        Maximum time for the solver to run in seconds
    bounds : ArrayLike, default=[0.0, np.inf]
        Bounds for the variables
    """

    name = ClassVar[str]
    short_name = ClassVar[str]

    # properties
    max_time: float
    bounds: ArrayLike

    def __init__(self):
        self.max_time = 3600
        self.bounds = [0.0, np.inf]

    def __repr__(self) -> str:
        return f"Solver {self.name} ({self.short_name})"

    @abstractmethod
    def solve(self, x0: ArrayLike) -> tuple[np.ndarray, dict]:
        """
        Interface method to solve the problem.

        Parameters
        ----------
        x0 : ArrayLike
            Initial guess for the solution

        Returns
        -------
        tuple[np.ndarray, dict]
            Solution vector and additional information as dictionary
        """


class NonLinearOptimizer(SolverBase):
    """
    Non-Linear Optimization Solver Base Class.

    Attributes
    ----------
    max_iter : int
        Maximum number of iterations
    abs_obj_tol : float
        Absolute objective tolerance
    objective : Callable
        Objective function handle
    gradient : Callable
        Gradient function handle
    hessian : Callable, default=None
        Hessian function handle
    constraints : Callable, default=None
        Constraints function handle
    constraints_jac : Callable, default=None
        Constraints Jacobian function handle
    supply_iter_func : bool
        Whether to supply an iteration callback function
    """

    max_iter: int
    abs_obj_tol: float

    objective: Callable
    gradient: Callable
    hessian: Callable
    constraints: Callable
    constraints_jac: Callable

    supply_iter_func: bool

    def __init__(self):
        super().__init__()
        self.max_iter = 500
        self.abs_obj_tol = 1e-6

        self.objective = None
        self.gradient = None
        self.hessian = None
        self.constraints = None
        self.constraints_jac = None

    def iter_func(self, *args, **kwargs) -> bool:
        """
        Get or set solver information as iteration callback.

        Agnostic signature with *args and **kwargs to be able to accomodate
        various solvers.

        Parameters
        ----------
        *args
            Additional arguments
        **kwargs
            Additional keyword arguments

        Returns
        -------
        bool
            Whether to continue the optimization
        """
        return True



from abc import ABC, abstractmethod
from typing import Union, ClassVar
import warnings
import logging

import numpy as np

# from numpy.typing import ArrayLike

from pyRadPlan.plan import Plan, validate_pln
from pyRadPlan.ct import CT, validate_ct
from pyRadPlan.cst import StructureSet, validate_cst
from pyRadPlan.stf import SteeringInformation, validate_stf
from pyRadPlan.dij import Dij, validate_dij
from pyRadPlan.scenarios import ScenarioModel
from pyRadPlan.quantities import FluenceDependentQuantity, get_quantity

from ..objectives import get_objective
from ..solvers import get_available_solvers, get_solver, SolverBase


logger = logging.getLogger(__name__)


class PlanningProblem(ABC):
    """
    Abstrac class for all planning problems.

    Parameters
    ----------
    pln : Union[Plan, dict], optional
        Plan object or dictionary to initialize the problem with.

    Attributes
    ----------
    short_name : ClassVar[str]
        Short name of the optimization problem.
    name : ClassVar[str]
        Name of the optimization problem.
    apply_overlap : bool, default=True
        Whether to apply overlap priorities to the StructureSet
    solver : Union[str, dict, SolverBase], default="ipopt"
        The solver to use for optimization.
    """

    # Constant, Abstract properties are realized as ClassVars
    short_name: ClassVar[str]
    name: ClassVar[str]
    possible_radiation_modes: list[str] = ["photons", "protons", "helium", "carbon", "oxygen"]

    apply_overlap: bool
    solver: Union[str, dict, SolverBase]

    # Private properties
    _ct: CT
    _cst: StructureSet
    _stf: SteeringInformation
    _dij: Dij
    _mult_scen: ScenarioModel

    _objective_list: list
    _constraint_list: list

    _quantities: list[FluenceDependentQuantity]
    _q_cache_index: list[int]
    _objectives_per_quantity: dict[str, int]

    def __init__(self, pln: Union[Plan, dict] = None):
        self._scenario_model = None

        self.solver = "ipopt"
        self.apply_overlap = True

        if pln is not None:
            pln = validate_pln(pln)
            self.assign_properties_from_pln(pln)

        solvers = get_available_solvers()
        if self.solver not in solvers:
            solver_names = list(solvers.keys())

            if len(solver_names) == 0:
                raise ValueError("No solver found!")

            warnings.warn(
                f"Solver {self.solver} not available. Choose from {solver_names}"
                ", and we will choose the first available one for you!"
            )

            self.solver = solver_names[0]

    def assign_properties_from_pln(self, pln: Plan, warn_when_property_changed: bool = False):
        """
        Assign properties from a Plan object to the Planning Problem.
        This function will check if a property exists for the PlanningProblem
        and, if yes, set it.

        Parameters
        ----------
        pln : Plan
            The Plan object to assign properties from.
        warn_when_property_changed : bool
            Whether to warn when properties are changed.
        """

        # Set Scenario Model
        self._mult_scen = pln.mult_scen

        # Assign Biologival Model
        if hasattr(pln, "bio_param"):
            self.bio_param = pln.bio_param  # TODO: No bio_param yet

        if not isinstance(warn_when_property_changed, bool):
            warn_when_property_changed = False

        # Overwrite default properties within the opti_prob
        # with the ones given in the prop_opt dict
        if hasattr(pln, "prop_opt") and isinstance(
            pln.prop_opt, dict
        ):  # TODO: This is not tested yet
            prop_dict = pln.prop_opt
            if (
                "opti_prob" in prop_dict
                and prop_dict["opti_prob"]
                and prop_dict["opti_prob"] != self.short_name
            ):
                raise ValueError(
                    f"Inconsistent dose opti_probs given! pln asks for '{prop_dict['opti_prob']}'"
                    f", but you are using '{self.short_name}'!"
                )
            prop_dict.pop("opti_prob", None)
        else:
            prop_dict = {}

        fields = prop_dict.keys()

        # Set up warning message
        if warn_when_property_changed:
            warning_msg = "Property in Optimization Problem overwritten from pln.prop_opt"
        else:
            warning_msg = None

        for field in fields:
            if not hasattr(self, field):
                warnings.warn(f"Property {field} not found in Problem!")
            elif warn_when_property_changed and warning_msg:
                logger.warning(warning_msg + f": {field}")

            setattr(self, field, prop_dict[field])

    @abstractmethod
    def _solve(self) -> tuple[np.ndarray, dict]:
        """Solve the planning problem."""

    def _initialize(self):
        """Initialize the data for the planning problem."""

        # resampling to dose-grid
        self._ct = self._ct.resample_to_grid(self._dij.dose_grid)

        # apply overlap priorities
        if self.apply_overlap:
            self._cst = self._cst.apply_overlap_priorities().resample_on_new_ct(self._ct)
        else:
            self._cst = self._cst.resample_on_new_ct(self._ct)

        # sanitize objectives and constraints and manage required quantities
        objectives = []
        quantity_ids = []
        for voi in self._cst.vois:
            if len(voi.objectives) > 0:
                # get the index list
                cube_ix = voi.indices_numpy
                objs = [get_objective(obj) for obj in voi.objectives]

                objectives.append((cube_ix, objs))

                quantity_ids.extend([obj.quantity for obj in objs])

        self._objective_list = objectives

        # unique quantities
        quantity_ids = list(set(quantity_ids))
        # get the quantities and check if they are fluence dependent
        quantities = [get_quantity(qid) for qid in quantity_ids]
        for q in quantities:
            if not issubclass(q, FluenceDependentQuantity):
                raise ValueError(
                    f"Quantity {q} is not fluence dependent! Currently only fluence dependent "
                    "quantities can be used in inverse planning!"
                )

        # TODO: manage scenarios

        # Manage quantites by getting them from the objective quantities
        self._quantities = [q(self._dij) for q in quantities]

        # obtain cache info to match quantities with objectives
        self._q_cache_index = []
        self._objectives_per_quantity = {q.identifier: [] for q in self._quantities}
        obj_ix = 0
        for obj_info in self._objective_list:
            for obj in obj_info[1]:
                for q in self._quantities:
                    if q.identifier == obj.quantity:
                        self._q_cache_index.append(
                            len(self._objectives_per_quantity[q.identifier])
                        )
                        self._objectives_per_quantity[q.identifier].append(obj_ix)
                    obj_ix += 1

        # Alternative code idea when storing tuples
        # quantity_obj_info = []
        # for q in quantities:
        #     q_instance = q(self._dij)

        #     # find objective indices with quantity]
        #     obj_ixs = []
        #     for ix, obj in enumerate(self._objective_list):
        #         if q_instance.identifier == obj.quantity:
        #             obj_ixs.append(ix)
        #     quantity_obj_info.append((q_instance, obj_ixs))
        # self._quantities = quantity_obj_info

        # set solver options
        self.solver = get_solver(self.solver)

        # initial point

    def solve(
        self,
        ct: Union[CT, dict],
        cst: Union[StructureSet, dict],
        stf: Union[SteeringInformation, dict],
        dij: Union[Dij, dict],
    ) -> tuple[np.ndarray, dict]:
        """
        Solves the planning problem.

        Will perform initialization & validation and call the desired Solver.

        Parameters
        ----------
        ct : Union[CT, dict]
            The CT object or compatible dictionary.
        cst : Union[StructureSet, dict]
            The StructureSet object or compatible dictionary.
        stf : Union[SteeringInformation, dict]
            The SteeringInformation object or compatible dictionary.
        dij : Union[Dij, dict]
            The Dij object or compatible dictionary.

        Returns
        -------
        tuple[np.ndarray,dict]
            The optimized result and additional solver-specific result information as dictionary.
        """

        self._ct = validate_ct(ct)
        self._cst = validate_cst(cst)
        self._stf = validate_stf(stf)
        self._dij = validate_dij(dij)

        self._initialize()
        return self._solve()


class NonLinearPlanningProblem(PlanningProblem):
    """Abstract Class for all Treatment Planning Problems."""

    @abstractmethod
    def _objective_functions(self, x: np.ndarray) -> np.ndarray:
        """Define the objective functions."""

    @abstractmethod
    def _objective_jacobian(self, x: np.ndarray) -> np.ndarray:
        """Define the objective jacobian."""

    def _objective_hessian(self, x: np.ndarray) -> np.ndarray:
        """Define the objective hessian."""
        return {}

    def _constraint_functions(self, x: np.ndarray) -> np.ndarray:
        """Define the constraint functions."""
        return None

    def _constraint_jacobian(self, x: np.ndarray) -> np.ndarray:
        """Define the constraint jacobian."""
        return None

    def _constraint_jacobian_structure(self) -> np.ndarray:
        """Define the constraint jacobian structure."""
        return None

    def _variable_bounds(self, x: np.ndarray) -> np.ndarray:
        """Define the variable bounds."""
        return np.array([0.0, np.inf], dtype=np.float64)



from abc import ABC, abstractmethod
from typing import ClassVar

import numpy as np
from numpy.typing import NDArray, ArrayLike
import pint

from pyRadPlan.dij import Dij, validate_dij

ureg = pint.UnitRegistry()


class RTQuantity(ABC):
    name: ClassVar[str]
    identifier: ClassVar[str]
    unit: ClassVar[pint.Unit]
    dim: ClassVar[int]  # To differentiate between scalar and vector quantities

    scenarios: NDArray[np.int64]  # Scenarios the quantity is calculated / defined for

    def __init__(self, scenarios=None):
        if scenarios is None:
            scenarios = [0]
        self.scenarios = np.asarray(scenarios, dtype=np.int64)


class FluenceDependentQuantity(RTQuantity, ABC):
    """Base class for quantities that depend on fluence distributions."""

    def __init__(self, dij: Dij, **kwargs):
        super().__init__(**kwargs)
        self._dij = validate_dij(dij)

        # Initialize Caches
        # Fluence cache for forward calculation
        self._w_cache = np.nan * np.ones((self._dij.total_num_of_bixels), dtype=np.float64)
        # Fluence cache for derivative calculation
        self._w_grad_cache = self._w_cache.copy()
        # Quantity vector cache
        self._q_cache = np.empty_like(getattr(self._dij, self.identifier), dtype=object)
        self._qgrad_cache = np.empty_like(self._q_cache)

    def __call__(self, fluence: ArrayLike) -> NDArray:
        """
        Make the quantity callable by calling the compute method.

        Parameters
        ----------
        fluence : ArrayLike
            Fluence vector.

        Returns
        -------
        NDArray
            Quantity vector.
        """

        return self.compute(fluence)

    def compute(self, fluence: ArrayLike) -> NDArray:
        """
        Forward calculation of the quantity from the fluence.

        Parameters
        ----------
        fluence : ArrayLike
            Fluence vector.

        Returns
        -------
        NDArray
            Quantity vector.
        """

        _fluence = np.asarray(fluence)

        if not np.array_equal(self._w_cache, _fluence):
            self._w_cache = _fluence.copy()
            self._compute_quantity_cache()

        return self._q_cache

    def compute_chain_derivative(self, d_quantity: ArrayLike, fluence: ArrayLike) -> NDArray:
        """
        Fluence Derivative of the quantity w.r.t. to the quantity derivative.

        Parameters
        ----------
        d_quantity : ArrayLike
            Derivative of w.r.t. to the quantity.
        fluence : ArrayLike
            Fluence vector.

        Returns
        -------
        NDArray
            Derivative of the quantity w.r.t. the fluence.
        """

        _d_quantity = np.asarray(d_quantity)
        _fluence = np.asarray(fluence)

        if not np.array_equal(self._w_grad_cache, _d_quantity):
            self._w_grad_cache = _fluence.copy()
            self._compute_chain_derivative_cache(_d_quantity)

        return self._qgrad_cache

    def _compute_quantity_cache(self):
        """
        Protected function to compute the quantity from the fluence and write it into the cache.

        Parameters
        ----------
        fluence : NDArray
            Fluence distribution.
        """

        for scenario_index in self.scenarios:
            self._q_cache.flat[scenario_index] = self._compute_quantity_single_scenario_from_cache(
                scenario_index
            )

    def _compute_chain_derivative_cache(self, d_quantity: NDArray) -> NDArray:
        """
        Protected interface for calculating the fluence derivative from quantity derivative.

        Parameters
        ----------
        d_quantity : NDArray
            Derivative w.r.t. to the quantity.

        Returns
        -------
        NDArray
            Derivative of the quantity w.r.t. the fluence.
        """

        for scenario_index in self.scenarios:
            self._qgrad_cache.flat[scenario_index] = (
                self._compute_chained_fluence_gradient_single_scenario_from_cache(
                    d_quantity, scenario_index
                )
            )

    @abstractmethod
    def _compute_quantity_single_scenario_from_cache(self, scenario_index: int) -> NDArray:
        """
        Calculate the quantity in a specific scenario.

        Parameters
        ----------
        scenario_index : int
            Scenario index.

        Returns
        -------
        NDArray
            Quantity in the scenario.
        """

    @abstractmethod
    def _compute_chained_fluence_gradient_single_scenario_from_cache(
        self, d_quantity: NDArray, scenario_index: int
    ) -> NDArray:
        """
        Calculate the derivative of the quantity w.r.t. the fluence in a specific scenario.

        Parameters
        ----------
        d_quantity : NDArray
            Derivative w.r.t. to the quantity.
        scenario_index : int
            Scenario index.

        Returns
        -------
        NDArray
            Derivative of the quantity w.r.t. the fluence in the scenario.
        """



"""pyRadPlan is a Python package for radiation therapy treatment planning."""  # noqa: N999

from importlib.metadata import version, PackageNotFoundError
import logging

from .plan._plans import Plan, validate_pln, IonPlan, PhotonPlan
from .ct._ct import CT, validate_ct
from .cst._cst import StructureSet, validate_cst
from .stf._generate_stf import generate_stf
from .stf import SteeringInformation, validate_stf
from .dose._calc_dose import calc_dose_influence, calc_dose_forward
from .optimization._fluence_optimization import fluence_optimization
from .analysis._dvh import DVH, DVHCollection
from .visualization import plot_slice
from .io import load_patient, load_tg119

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    # package is not installed
    pass

# Logging is not exposed by default and needs to be configured by the user.
logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "__version__",
    "Plan",
    "IonPlan",
    "PhotonPlan",
    "validate_pln",
    "CT",
    "validate_ct",
    "StructureSet",
    "validate_cst",
    "generate_stf",
    "calc_dose_influence",
    "calc_dose_forward",
    "fluence_optimization",
    "DVH",
    "DVHCollection",
    "SteeringInformation",
    "validate_stf",
    "plot_slice",
    "load_patient",
    "load_tg119",
]



from ._base import FluenceDependentQuantity, RTQuantity
from ._dose import Dose
from ._let_x_dose import LETxDose

QUANTITIES = {Dose.identifier: Dose, LETxDose.identifier: LETxDose}


def get_available_quantities() -> dict[str, RTQuantity]:
    """
    Obtain the available quantities in planning.

    Returns
    -------
    dict
        Dictionary with the available quantities
    """
    return QUANTITIES


def get_quantity(identifier: str) -> RTQuantity:
    """
    Obtain the quantity from name.

    Parameters
    ----------
    identifier : str
        The identifier of the quantity

    Returns
    -------
    RTQuantity
        The quantity
    """
    return QUANTITIES[identifier]


__all__ = [
    "FluenceDependentQuantity",
    "Dose",
    "RTQuantity",
    "get_available_quantities",
    "get_quantity",
]



from typing import Any, Union
from typing_extensions import Self
from pydantic import (
    Field,
    field_validator,
    computed_field,
)
from numpydantic import NDArray, Shape
import numpy as np
import SimpleITK as sitk

from .datamodel import PyRadPlanBaseModel


class Grid(PyRadPlanBaseModel):
    """
    Class representing image grids in the LPS world system.

    Attributes
    ----------
    resolution : dict[str, float]
        The resolution of the grid in the x, y, and z directions.
    dimensions : tuple[int, int, int]
        The dimensions of the grid in the x, y, and z directions.
    origin : np.ndarray
        The origin of the grid in the LPS world system.
    direction : np.ndarray
        The direction cosines of the grid in the LPS world system.
    """

    resolution: dict[str, float]
    dimensions: tuple[int, int, int]
    origin: NDArray[Shape["3"], np.floating] = Field(
        default=np.array([0.0, 0.0, 0.0], dtype=np.float64), alias="cubeCoordOffset"
    )
    direction: NDArray[Shape["3,3"], np.floating] = Field(default=np.eye(3, dtype=np.float64))

    @computed_field(alias="numOfVoxels")
    @property
    def num_voxels(self) -> int:
        """Number of voxels in the grid."""
        return int(np.prod(self.dimensions))

    @computed_field
    @property
    def x(self) -> float:
        """Return the x coordinates in the LPS world system."""
        x = np.arange(self.dimensions[0]) * self.resolution["x"] + self.origin[0]
        return x

    @computed_field
    @property
    def y(self) -> float:
        """Return the y coordinates in the LPS world system."""
        y = np.arange(self.dimensions[1]) * self.resolution["y"] + self.origin[1]
        return y

    @computed_field
    @property
    def z(self) -> float:
        """Return the z coordinates in the LPS world system."""
        z = np.arange(self.dimensions[2]) * self.resolution["z"] + self.origin[2]
        return z

    @property
    def resolution_vector(self) -> np.ndarray:
        """Return the resolution as a vector."""
        return np.array([self.resolution["x"], self.resolution["y"], self.resolution["z"]])

    @field_validator("resolution", mode="after")
    @classmethod
    def _check_resolution(cls, value: dict[str, float]) -> dict[str, float]:
        """Check if resolution has the correct structure and values."""
        # Check if resolution has the correct structure (dict with keys 'x', 'y', 'z')
        if not all(key in value for key in ["x", "y", "z"]):
            raise ValueError("resolution must have keys 'x', 'y', 'z'")

        # Check if all resolution values are positive floats. If not, try to cast them to float
        for _, v in value.items():
            if v <= 0:
                raise ValueError(f"resolution values must be positive floats, got {v}")

        return value

    @field_validator("dimensions")
    @classmethod
    def _check_dimensions(cls, value: tuple[int, int, int]) -> tuple[int, int, int]:
        """Check if dimensions has the correct structure and values."""
        # Check if dimensions has exactly 3 elements
        if len(value) != 3:
            raise ValueError("dimensions must have exactly 3 elements")

        # Check if all dimensions values are positive integers. If not, try to cast them to int
        for dim in value:
            try:
                tmpdim = int(dim)
            except ValueError:
                raise ValueError(f"dimension value could not be casted into int, got {dim}")

            if tmpdim <= 0:
                raise ValueError(f"dimension values must be positive integers, got {tmpdim}")

        return value

    @field_validator("origin", mode="before")
    @classmethod
    def _check_origin(cls, value: Any) -> Any:
        """Check if origin has the correct shape (3,) and values."""
        # Check if origin has the correct shape (3,)
        try:
            value = np.asarray(value, dtype=np.float64).reshape((3,))
        except ValueError as exc:
            raise ValueError("origin must be convertible to a 1D numpy array of length 3") from exc
        return value

    @field_validator("direction", mode="before")
    @classmethod
    def _check_direction(cls, value: Any) -> Any:
        """Check if direction has the correct shape (3x3)."""

        try:
            value = np.asarray(value, dtype=np.float64).reshape((3, 3))
        except ValueError as exc:
            raise ValueError("direction must be convertible to a 3x3 numpy matrix") from exc
        return value

    # TODO: validate for additional fields if, e.g., loaded from matRad

    def to_matrad(self, context: str = "mat-file") -> Any:
        grid_dict4matrad = super().to_matrad(context=context)
        grid_dict4matrad["dimensions"] = tuple(map(float, grid_dict4matrad["dimensions"]))
        grid_dict4matrad["numOfVoxels"] = float(grid_dict4matrad["numOfVoxels"])
        return grid_dict4matrad

    @classmethod
    def from_sitk_image(cls, sitk_image: sitk.Image) -> Self:
        """
        Create a Grid object from a SimpleITK image.

        Parameters
        ----------
        sitk_image : sitk.Image
            The SimpleITK image to create the Grid object from.

        Returns
        -------
        Grid
            The Grid object created from the SimpleITK image.
        """
        keys = ["x", "y", "z"]
        resolution = dict(zip(keys, sitk_image.GetSpacing()))
        dimensions = sitk_image.GetSize()
        origin = sitk_image.GetOrigin()
        direction = sitk_image.GetDirection()
        return cls(
            resolution=resolution, dimensions=dimensions, origin=origin, direction=direction
        )

    def resample(
        self,
        target_resolution: Union[
            dict[str, float], np.ndarray, tuple[float, float, float], list[float]
        ],
    ) -> Self:
        """
        Create a resampled grid covering the original grid in new resolution.

        Parameters
        ----------
        target_resolution : Union[dict[str,float], np.ndarray]
            The target resolution of the resampled grid.

        Returns
        -------
        Grid
            The resampled grid object.
        """

        if isinstance(target_resolution, dict):
            target_resolution = np.array(
                [target_resolution["x"], target_resolution["y"], target_resolution["z"]]
            )
        else:
            try:
                target_resolution = np.asarray(target_resolution).reshape(
                    3,
                )
            except ValueError as exc:
                raise ValueError(
                    "target_resolution must be convertible to an ndarray of shape (3,)"
                ) from exc

        # Calculate the new dimensions
        new_dimensions = np.ceil(
            np.array(self.dimensions) * self.resolution_vector / target_resolution
        ).astype(int)

        # Now calculate the width of the grid in the new resolution
        old_grid_size = np.array(self.dimensions) * self.resolution_vector
        new_grid_size = new_dimensions * target_resolution
        diff_size = new_grid_size - old_grid_size

        # find the origin of the new image such that it is centered above the old image. The origin
        # is in the center of the first voxel
        origin_shift = diff_size / 2.0

        # now we need to respect the origin and direction to correctly shift the new grid into
        # position
        origin_shift = np.matmul(self.direction, origin_shift)
        new_origin = self.origin - origin_shift

        return Grid(
            resolution=dict(zip(["x", "y", "z"], target_resolution)),
            dimensions=tuple(new_dimensions),
            origin=new_origin,
            direction=self.direction,
        )


if __name__ == "__main__":
    # Example usage
    try:
        grid_dict = {
            "resolution": {"x": 1, "y": 1, "z": 1},
            "dimensions": (10, 10, 10),
        }

        grid_information = Grid.model_validate(grid_dict)
        print(grid_information.to_matrad())
    except Exception as e:
        print("Error creating grid from dose_information:", e)



import pytest
import numpy as np
import SimpleITK as sitk
from pyRadPlan.cst import OAR, Target, HelperVOI, ExternalVOI, create_voi


def test_create_voi_no_args():
    with pytest.raises(ValueError):
        create_voi()


def test_target_constructor_empty_args():
    with pytest.raises(ValueError):
        Target()


def test_oar_constructor_empty_args():
    with pytest.raises(ValueError):
        OAR()


def test_helper_voi_constructor_empty_args():
    with pytest.raises(ValueError):
        HelperVOI()


def test_target_constructor_3d(generic_input_3d):
    name, ct, mask, alpha_x, beta_x = generic_input_3d
    target = Target(name=name, ct_image=ct, mask=mask)
    assert target.ct_image == ct
    assert isinstance(target.mask, sitk.Image)
    assert target.voi_type == "TARGET"
    assert target.mask.GetOrigin() == target.ct_image.cube_hu.GetOrigin()
    assert target.mask.GetSpacing() == target.ct_image.cube_hu.GetSpacing()
    assert target.mask.GetDirection() == target.ct_image.cube_hu.GetDirection()

    # test non default alpha_x and beta_x
    target_2 = Target(name=name, ct_image=ct, mask=mask, alpha_x=alpha_x, beta_x=beta_x)
    assert target_2.alpha_x == alpha_x
    assert target_2.beta_x == beta_x


def test_target_constructor_4d(generic_input_4d):
    name, ct, mask, alpha_x, beta_x = generic_input_4d
    target = Target(name=name, ct_image=ct, mask=mask)
    assert target.ct_image == ct
    assert isinstance(target.mask, sitk.Image)
    assert target.voi_type == "TARGET"

    # test non default alpha_x and beta_x
    target_2 = Target(name=name, ct_image=ct, mask=mask, alpha_x=alpha_x, beta_x=beta_x)
    assert target_2.alpha_x == alpha_x
    assert target_2.beta_x == beta_x


def test_target_constructor_4d_np(generic_input_4d):
    name, ct, mask, alpha_x, beta_x = generic_input_4d
    target = Target(name=name, ct_image=ct, mask=sitk.GetArrayFromImage(mask))
    assert target.ct_image == ct
    assert isinstance(target.mask, sitk.Image)
    assert target.voi_type == "TARGET"

    # test non default alpha_x and beta_x
    target_2 = Target(name=name, ct_image=ct, mask=mask, alpha_x=alpha_x, beta_x=beta_x)
    assert target_2.alpha_x == alpha_x
    assert target_2.beta_x == beta_x


def test_mixed_constructors(generic_input_3d, generic_input_4d):
    name_3d, ct_3d, mask_3d, _, _ = generic_input_3d
    name_4d, ct_4d, mask_4d, _, _ = generic_input_4d
    with pytest.raises(ValueError):
        target = Target(name=name_3d, ct_image=ct_3d, mask=mask_4d)

    with pytest.raises(ValueError):
        target = Target(name=name_4d, ct_image=ct_4d, mask=mask_3d)


def test_oar_constructor(generic_input_3d):
    name, ct, mask, alpha_x, beta_x = generic_input_3d
    oar = OAR(name=name, ct_image=ct, mask=mask)
    assert oar.ct_image == ct
    assert isinstance(oar.mask, sitk.Image)
    assert oar.voi_type == "OAR"

    # test non default alpha_x and beta_x
    oar_2 = OAR(name=name, ct_image=ct, mask=mask, alpha_x=alpha_x, beta_x=beta_x)
    assert oar_2.alpha_x == alpha_x
    assert oar_2.beta_x == beta_x


def test_helper_voi_constructor(generic_input_3d):
    name, ct, mask, alpha_x, beta_x = generic_input_3d
    helper_voi = HelperVOI(name=name, ct_image=ct, mask=mask)
    assert helper_voi.ct_image == ct
    assert isinstance(helper_voi.mask, sitk.Image)
    assert helper_voi.voi_type == "HELPER"

    # test non default alpha_x and beta_x
    helper_voi_2 = HelperVOI(name=name, ct_image=ct, mask=mask, alpha_x=alpha_x, beta_x=beta_x)
    assert helper_voi_2.alpha_x == alpha_x
    assert helper_voi_2.beta_x == beta_x


def tet_external_voi_constructor(generic_input_3d):
    name, ct, mask, alpha_x, beta_x = generic_input_3d
    helper_voi = ExternalVOI(name=name, ct_image=ct, mask=mask)
    assert helper_voi.ct_image == ct
    assert isinstance(helper_voi.mask, sitk.Image)
    assert helper_voi.voi_type == "HELPER"

    # test non default alpha_x and beta_x
    helper_voi_2 = ExternalVOI(name=name, ct_image=ct, mask=mask, alpha_x=alpha_x, beta_x=beta_x)
    assert helper_voi_2.alpha_x == alpha_x
    assert helper_voi_2.beta_x == beta_x


def test_voi_idx_wrong_shape(generic_input_3d):
    name, ct, _, _, _ = generic_input_3d
    with pytest.raises(ValueError):
        Target(name=name, ct_image=ct, mask=np.array([[1], [1], [1]]))


def test_voi_idx_wrong_dim(generic_input_3d):
    name, ct, _, _, _ = generic_input_3d
    with pytest.raises(ValueError):
        Target(name=name, ct_image=ct, mask=np.array([1]))

    with pytest.raises(ValueError):
        Target(name=name, ct_image=ct, mask=np.array([1] * 5)[:, None])


def test_voi_idx_non_int_idx(generic_input_3d):
    name, ct, _, _, _ = generic_input_3d
    with pytest.raises(ValueError):
        Target(name=name, ct_image=ct, mask=[[1.2, 3, 4]])


def test_voi_idx_wrong_dtype(generic_input_3d):
    name, ct, mask, _, _ = generic_input_3d
    mask_float = mask.astype("float32")
    mask_float = sitk.GetImageFromArray(mask_float)
    with pytest.raises(ValueError):
        Target(name=name, ct_image=ct, mask=mask_float)

    mask = np.zeros_like(mask) * 3.0
    with pytest.raises(ValueError):
        Target(name=name, ct_image=ct, mask=mask)


def test_voi_idx_wrong_input(generic_input_3d):
    name, ct, _, _, _ = generic_input_3d
    with pytest.raises(ValueError):
        Target(name=name, ct_image=ct, mask=[[2], [3], [4]])


def test_create_voi_target_from_dict(generic_input_3d):
    name, ct, mask, _, _ = generic_input_3d
    voi = create_voi(data={"voi_type": "TARGET", "name": name, "ct_image": ct, "mask": mask})
    assert isinstance(voi, Target)
    assert voi.ct_image == ct
    assert isinstance(voi.mask, sitk.Image)
    assert voi.voi_type == "TARGET"


def test_create_voi_oar_from_dict(generic_input_3d):
    name, ct, mask, _, _ = generic_input_3d
    voi = create_voi(data={"voi_type": "OAR", "name": name, "ct_image": ct, "mask": mask})
    assert isinstance(voi, OAR)
    assert voi.ct_image == ct
    assert isinstance(voi.mask, sitk.Image)
    assert voi.voi_type == "OAR"


def test_create_voi_helper_from_dict(generic_input_3d):
    name, ct, mask, _, _ = generic_input_3d
    voi = create_voi(data={"voi_type": "HELPER", "name": name, "ct_image": ct, "mask": mask})
    assert isinstance(voi, HelperVOI)
    assert voi.ct_image == ct
    assert isinstance(voi.mask, sitk.Image)
    assert voi.voi_type == "HELPER"


def test_create_voi_invalid_voi_type(generic_input_3d):
    name, ct, mask, _, _ = generic_input_3d
    with pytest.raises(ValueError):
        create_voi(data={"voi_type": "invalid", "name": name, "ct_image": ct, "mask": mask})


def test_create_voi_no_voi_type(generic_input_3d):
    name, ct, mask, _, _ = generic_input_3d
    with pytest.raises(ValueError):
        create_voi(data={"name": name, "ct_image": ct, "mask": mask})


def test_create_voi_no_data():
    with pytest.raises(ValueError):
        create_voi()


def test_create_voi_from_VOI(generic_input_3d):
    name, ct, mask, _, _ = generic_input_3d
    voi = create_voi(data={"voi_type": "HELPER", "name": name, "ct_image": ct, "mask": mask})
    voi_2 = create_voi(voi)
    assert voi == voi_2


def test_create_target_from_kwargs(generic_input_3d):
    name, ct, mask, _, _ = generic_input_3d
    voi = create_voi(voi_type="TARGET", name=name, ct_image=ct, mask=mask)
    assert isinstance(voi, Target)
    assert voi.ct_image == ct
    assert isinstance(voi.mask, sitk.Image)
    assert voi.voi_type == "TARGET"


def test_create_oar_from_kwargs(generic_input_3d):
    name, ct, mask, _, _ = generic_input_3d
    voi = create_voi(voi_type="OAR", name=name, ct_image=ct, mask=mask)
    assert isinstance(voi, OAR)
    assert voi.ct_image == ct
    assert isinstance(voi.mask, sitk.Image)
    assert voi.voi_type == "OAR"


def test_create_helper_from_kwargs(generic_input_3d):
    name, ct, mask, _, _ = generic_input_3d
    voi = create_voi(voi_type="HELPER", name=name, ct_image=ct, mask=mask)
    assert isinstance(voi, HelperVOI)
    assert voi.ct_image == ct
    assert isinstance(voi.mask, sitk.Image)
    assert voi.voi_type == "HELPER"


def test_create_voi_invalid_voi_type_kwargs(generic_input_3d):
    name, ct, mask, _, _ = generic_input_3d
    with pytest.raises(ValueError):
        create_voi(voi_type="invalid", name=name, ct_image=ct, mask=mask)


def test_create_voi_no_voi_type_kwargs(generic_input_3d):
    name, ct, mask, _, _ = generic_input_3d
    with pytest.raises(ValueError):
        create_voi(name=name, ct_image=ct, mask=mask)


def test_voi_indices(generic_input_3d, generic_input_4d):
    name_3d, ct_3d, mask_3d, _, _ = generic_input_3d
    name_4d, ct_4d, mask_4d, _, _ = generic_input_4d
    voi_3d = create_voi(voi_type="TARGET", name=name_3d, ct_image=ct_3d, mask=mask_3d)
    voi_4d = create_voi(voi_type="TARGET", name=name_4d, ct_image=ct_4d, mask=mask_4d)
    assert (voi_3d.indices == np.array([5000])).all()
    assert (sitk.GetArrayViewFromImage(voi_3d.mask).ravel(order="F")[voi_3d.indices] == 1).all()
    assert (voi_4d.indices == np.array([103, 10000])).all()
    assert (sitk.GetArrayViewFromImage(voi_4d.mask).ravel(order="F")[voi_4d.indices] == 1).all()


def test_voi_indices_np(generic_input_3d, generic_input_4d):
    name_3d, ct_3d, mask_3d, _, _ = generic_input_3d
    name_4d, ct_4d, mask_4d, _, _ = generic_input_4d
    voi_3d = create_voi(voi_type="TARGET", name=name_3d, ct_image=ct_3d, mask=mask_3d)
    voi_4d = create_voi(voi_type="TARGET", name=name_4d, ct_image=ct_4d, mask=mask_4d)
    assert (voi_3d.indices_numpy == np.array([1])).all()
    assert (
        sitk.GetArrayViewFromImage(voi_3d.mask).ravel(order="C")[voi_3d.indices_numpy] == 1
    ).all()
    assert (voi_4d.indices_numpy == np.array([1, 510100])).all()
    assert (
        sitk.GetArrayViewFromImage(voi_4d.mask).ravel(order="C")[voi_4d.indices_numpy] == 1
    ).all()


def test_voi_get_indices_by_order(generic_input_3d, generic_input_4d):
    name_3d, ct_3d, mask_3d, _, _ = generic_input_3d
    name_4d, ct_4d, mask_4d, _, _ = generic_input_4d
    voi_3d = create_voi(voi_type="TARGET", name=name_3d, ct_image=ct_3d, mask=mask_3d)
    voi_4d = create_voi(voi_type="TARGET", name=name_4d, ct_image=ct_4d, mask=mask_4d)
    assert (voi_3d.get_indices(order="numpy") == voi_3d.indices_numpy).all()
    assert (voi_4d.get_indices(order="numpy") == voi_4d.indices_numpy).all()
    assert (voi_3d.get_indices(order="sitk") == voi_3d.indices).all()
    assert (voi_4d.get_indices(order="sitk") == voi_4d.indices).all()


def test_voi_numpy_mask(generic_input_3d, generic_input_4d):
    name_3d, ct_3d, mask_3d, _, _ = generic_input_3d
    name_4d, ct_4d, mask_4d, _, _ = generic_input_4d
    voi_3d = create_voi(voi_type="TARGET", name=name_3d, ct_image=ct_3d, mask=mask_3d)
    voi_4d = create_voi(voi_type="TARGET", name=name_4d, ct_image=ct_4d, mask=mask_4d)
    assert (voi_3d._numpy_mask == mask_3d).all()
    assert (voi_4d._numpy_mask == sitk.GetArrayViewFromImage(mask_4d)).all()


def test_scenario_indices(generic_input_3d, generic_input_4d):
    name_3d, ct_3d, mask_3d, _, _ = generic_input_3d
    name_4d, ct_4d, mask_4d, _, _ = generic_input_4d
    voi_3d = create_voi(voi_type="TARGET", name=name_3d, ct_image=ct_3d, mask=mask_3d)
    voi_4d = create_voi(voi_type="TARGET", name=name_4d, ct_image=ct_4d, mask=mask_4d)
    assert (voi_3d.scenario_indices() == np.array([1])).all()
    assert (voi_3d.scenario_indices("sitk") == np.array([5000])).all()
    assert voi_4d.scenario_indices() == [np.array([1]), np.array([10100])]
    assert voi_4d.scenario_indices("sitk") == [np.array([5000]), np.array([51])]
    with pytest.raises(ValueError):
        voi_3d.scenario_indices("invalid")
    with pytest.raises(ValueError):
        voi_4d.scenario_indices("invalid")


def test_masked_ct(generic_input_3d, generic_input_4d):
    name_3d, ct_3d, mask_3d, _, _ = generic_input_3d
    name_4d, ct_4d, mask_4d, _, _ = generic_input_4d
    voi_3d = create_voi(voi_type="TARGET", name=name_3d, ct_image=ct_3d, mask=mask_3d)
    voi_4d = create_voi(voi_type="TARGET", name=name_4d, ct_image=ct_4d, mask=mask_4d)
    with pytest.raises(ValueError):
        voi_3d.masked_ct("invalid")
    with pytest.raises(ValueError):
        voi_4d.masked_ct("invalid")

    masked_ct_3d_sitk = voi_3d.masked_ct("sitk")
    assert isinstance(masked_ct_3d_sitk, sitk.Image)
    assert (sitk.GetArrayFromImage(masked_ct_3d_sitk) == mask_3d.astype(np.float32) * 1000).all()

    masked_ct_3d_np = voi_3d.masked_ct("numpy")
    assert (masked_ct_3d_np == mask_3d.astype(np.float32) * 1000).all()

    masked_ct_4d_sitk = voi_4d.masked_ct("sitk")
    assert isinstance(masked_ct_4d_sitk, sitk.Image)
    assert (
        sitk.GetArrayFromImage(masked_ct_4d_sitk)
        == sitk.GetArrayViewFromImage(mask_4d).astype(np.float32) * 1000
    ).all()

    masked_ct_4d_np = voi_4d.masked_ct("numpy")
    assert (masked_ct_4d_np == sitk.GetArrayViewFromImage(mask_4d).astype(np.float32) * 1000).all()


def test_scenario_ct_data(generic_input_3d, generic_input_4d):
    name_3d, ct_3d, mask_3d, _, _ = generic_input_3d
    name_4d, ct_4d, mask_4d, _, _ = generic_input_4d
    voi_3d = create_voi(voi_type="TARGET", name=name_3d, ct_image=ct_3d, mask=mask_3d)
    voi_4d = create_voi(voi_type="TARGET", name=name_4d, ct_image=ct_4d, mask=mask_4d)
    assert isinstance(voi_4d.scenario_ct_data, list)
    assert len(voi_4d.scenario_ct_data) == 2
    assert (voi_3d.scenario_ct_data == np.array([1000])).all()
    assert (voi_4d.scenario_ct_data[0] == np.array([1000])).all()
    assert (voi_4d.scenario_ct_data[1] == np.array([1000])).all()


def test_create_target_camel_case(generic_input_3d):
    name, ct, mask, alpha_x, beta_x = generic_input_3d

    voi = create_voi(
        data={
            "voiType": "TARGET",
            "name": name,
            "ctImage": ct,
            "mask": mask,
            "alphaX": alpha_x,
            "betaX": beta_x,
        }
    )
    assert isinstance(voi, Target)
    assert voi.ct_image == ct
    assert isinstance(voi.mask, sitk.Image)
    assert voi.voi_type == "TARGET"
    assert voi.alpha_x == alpha_x
    assert voi.beta_x == beta_x


def test_create_oar_camel_case(generic_input_3d):
    name, ct, mask, alpha_x, beta_x = generic_input_3d
    voi = create_voi(
        data={
            "voiType": "OAR",
            "name": name,
            "ctImage": ct,
            "mask": mask,
            "alphaX": alpha_x,
            "betaX": beta_x,
        }
    )
    assert isinstance(voi, OAR)
    assert voi.ct_image == ct
    assert isinstance(voi.mask, sitk.Image)
    assert voi.voi_type == "OAR"
    assert voi.alpha_x == alpha_x
    assert voi.beta_x == beta_x


def test_create_helper_camel_case(generic_input_3d):
    name, ct, mask, alpha_x, beta_x = generic_input_3d
    voi = create_voi(
        data={
            "voiType": "HELPER",
            "name": name,
            "ctImage": ct,
            "mask": mask,
            "alphaX": alpha_x,
            "betaX": beta_x,
        }
    )
    assert isinstance(voi, HelperVOI)
    assert voi.ct_image == ct
    assert isinstance(voi.mask, sitk.Image)
    assert voi.voi_type == "HELPER"
    assert voi.alpha_x == alpha_x
    assert voi.beta_x == beta_x


def test_create_voi_invalid_voi_type_camel_case(generic_input_3d):
    name, ct, mask, _, _ = generic_input_3d
    with pytest.raises(ValueError):
        create_voi(voi_type="invalid", name=name, ctImage=ct, mask=mask)


# def test_valildate_voi(generic_input_3d):
#     name, ct, mask, _, _=  generic_input_3d
#     voi = create_voi(voi_type="TARGET", name=name, ct_image=ct, mask=mask)
#     voivalidate_voi()
#     voi_2 = create_voi(voi_type="OAR", name=name, ct_image=ct, mask=mask)
#     voi_2.validate_voi()
#     voi_3 = create_voi(voi_type="HELPER", name=name, ct_image=ct, mask=mask)
#     voi_3.validate_voi()



"""Geometry functions for LPS system."""

import numpy as np


def get_gantry_rotation_matrix(gantry_angle):
    """
    Calculate the rotation matrix for gantry.

    Represents an active, counter-clockwise rotation Matrix for Gantry around z
    with pre-multiplication of the matrix (R*x).

    Parameters
    ----------
    gantry_angle : float
        The angle of gantry rotation in degrees.

    Returns
    -------
    numpy.ndarray
        The rotation matrix.

    Notes
    -----
    Gantry rotation is physically an active rotation of a beam
    vector around the target / isocenterin the patient coordinate
    system

    The LPS system is right-handed, where the gantry rotates counter-clockwise around the z-axis
    and the couch rotates counter-clockwise around the y-axis.
    The gantry rotation is an active rotation of a beam vector around the target/isocenter
    in the patient coordinate system.
    """

    gantry_angle = np.deg2rad(gantry_angle)

    r_gantry = np.array(
        [
            [np.cos(gantry_angle), -np.sin(gantry_angle), 0],
            [np.sin(gantry_angle), np.cos(gantry_angle), 0],
            [0, 0, 1],
        ]
    )

    return r_gantry


def get_couch_rotation_matrix(couch_angle):
    """
    Calculate the rotation matrix for the couch.

    Parameters
    ----------
    couch_angle : float
        The angle of couch rotation in degrees.

    Returns
    -------
    numpy.ndarray
        The rotation matrix.

    Notes
    -----
    The LPS system is right-handed, where the gantry rotates counter-clockwise around the z-axis
    and the couch rotates counter-clockwise around the y-axis.
    The gantry rotation is an active rotation of a beam vector around the target/isocenter
    in the patient coordinate system.
    """

    couch_angle = np.deg2rad(couch_angle)

    r_gantry = np.array(
        [
            [np.cos(couch_angle), 0.0, np.sin(couch_angle)],
            [0.0, 1.0, 0.0],
            [-np.sin(couch_angle), 0.0, np.cos(couch_angle)],
        ]
    )

    return r_gantry


def get_beam_rotation_matrix(gantry_angle, couch_angle):
    """
    Calculate the rotation matrix for gantry and couch angles.

    Represents active, counter-clockwise rotation for couch around y
    with pre-multiplication of the matrix (R*x)

    Parameters
    ----------
    gantry_angle : float
        The angle of gantry rotation in degrees.
    couch_angle : float
        The angle of couch rotation in degrees.

    Returns
    -------
    numpy.ndarray
        The rotation matrix.

    Notes
    -----
    Couch rotation is physically a passive rotation of the
    patient system around the beam target point / isocenter

    The LPS system is right-handed, where the gantry rotates counter-clockwise around the z-axis
    and the couch rotates counter-clockwise around the y-axis.
    The gantry rotation is an active rotation of a beam vector around the target/isocenter
    in the patient coordinate system.

    Examples
    --------
    >>> get_beam_rotation_matrix(90, 45)
    array([[-0.70710678, -0.70710678,  0.        ],
           [ 0.        ,  0.        ,  1.        ],
           [ 0.70710678, -0.70710678,  0.        ]])
    """

    r_gantry = get_gantry_rotation_matrix(gantry_angle)
    r_couch = get_couch_rotation_matrix(couch_angle)

    rot_mat = r_couch @ r_gantry

    return rot_mat



import pytest
from pyRadPlan import load_tg119
import SimpleITK as sitk
from pyRadPlan.analysis import DVHCollection, DVH
import numpy as np
from matplotlib import pyplot as plt


@pytest.fixture
def cst():
    # Get StructureSet from TG119
    return load_tg119()[1]


@pytest.fixture
def dose():
    # Create dose image using information from TG119
    ct, _ = load_tg119()
    dose_array = np.zeros(ct.size[::-1])
    dose_array[60:80, 80:120, 80:120] = 1

    # Create dose image
    dose_image = sitk.GetImageFromArray(dose_array)
    # Copy information from CT image
    dose_image.SetSpacing((ct.resolution["x"], ct.resolution["y"], ct.resolution["z"]))
    dose_image.SetOrigin((ct.origin[0], ct.origin[1], ct.origin[2]))
    dose_image.SetDirection(
        (
            ct.direction[0],
            ct.direction[1],
            ct.direction[2],
            ct.direction[3],
            ct.direction[4],
            ct.direction[5],
            ct.direction[6],
            ct.direction[7],
            ct.direction[8],
        )
    )
    return dose_image


def test_dvhcollection(cst, dose):
    dvh = DVHCollection.from_structure_set(cst=cst, dose=dose)
    assert len(dvh.dvhs) == len(cst.vois)


def test_dvh(cst, dose):
    n_points = 500
    dvh = DVH.compute(
        mask=cst.vois[0].mask, quantity=dose, name=cst.vois[0].name, num_points=n_points
    )
    assert dvh.name == cst.vois[0].name
    assert dvh.bin_edges.shape == (n_points + 1,)
    assert dvh.bins.shape == (n_points,)
    assert np.all(dvh.bins == dvh.bin_edges[:-1])
    assert dvh.bin_centers.shape == (n_points,)
    assert np.all(dvh.bin_centers > dvh.bins)
    assert dvh.diff_volume.shape == (n_points,)
    assert dvh.cum_volume.shape == (n_points,)
    assert dvh.cumulative.shape == (2, n_points)
    assert dvh.differential.shape == (2, n_points)
    assert np.array_equal(dvh.cumulative[0], dvh.bins)
    assert np.array_equal(dvh.cumulative[1], dvh.cum_volume)
    assert np.array_equal(dvh.differential[0], dvh.bin_centers)
    assert np.array_equal(dvh.differential[1], dvh.diff_volume)

    assert isinstance(dvh.has_regular_bins, np.bool)
    assert dvh.get_vx(1.0) == np.interp(1.0, dvh.cumulative[0], dvh.cumulative[1])
    assert dvh.get_dy(50) > 0.0



import SimpleITK as sitk
import numpy as np

from pyRadPlan.dose import calc_dose_forward


# TODO: PhotonEngine does not yet have sub-sampling.
# this here only tests if it runs without error.
def test_photons(test_data_photons):
    pln, ct, cst, stf, dij, result = test_data_photons

    pln.prop_dose_calc["dosimetric_lateral_cutoff"] = 0.995
    pln.prop_dose_calc["lateral_model"] = "single"

    result_py = calc_dose_forward(ct, cst, stf, pln, weights=None)
    result_py = sitk.GetArrayFromImage(result_py["physical_dose"])

    result_matRad_rot = np.swapaxes(result["physicalDose"], 0, 1)

    # assert np.allclose(result_py, result_matRad_rot, atol=1e-4)

    # Keeping this for debugging:
    # plot_slice(
    #     ct=ct,
    #     cst=cst,
    #     overlay=result_matRad_rot,#-result_matRad_rot,
    #     view_slice=5,
    #     plane="axial",
    #     overlay_unit="Gy",
    #     plt_show = True,
    #     use_global_max = False,
    # )



"""Contains all custom exceptions for the stf module."""

from pyRadPlan.core import PyRadPlanError


class GeometryError(PyRadPlanError):
    """Defines an error in the geometry during generation of the stf."""

    def __init__(self, message: str):
        super().__init__(message)



"""
Ipopt solver for non-linear optimization problems.

Notes
-----
Not installed by default. Uses ipyopt because it provides linux wheels
"""

from ipyopt import Problem

import numpy as np
from numpy.typing import ArrayLike, NDArray

from ._base_solvers import NonLinearOptimizer


class OptimizerIpopt(NonLinearOptimizer):
    """
    IPOPT solver interface.

    Attributes
    ----------
    options : dict
        Options for IPOPT
    """

    name = "Interior Point Optimizer"
    short_name = "ipopt"

    options: dict[str]

    def __init__(self):
        self.result = None

        super().__init__()

        self.options = {
            "print_level": 5,
            "print_user_options": "no",
            "print_options_documentation": "no",
            "tol": 1e-10,
            "dual_inf_tol": 1e-4,
            "constr_viol_tol": 1e-4,
            "compl_inf_tol": 1e-4,
            "acceptable_iter": 5,
            "acceptable_tol": self.abs_obj_tol,
            "acceptable_constr_viol_tol": 1e-2,
            "acceptable_dual_inf_tol": 1e10,
            "acceptable_compl_inf_tol": 1e10,
            "acceptable_obj_change_tol": 1e-4,
            "max_iter": self.max_iter,
            "max_cpu_time": float(self.max_time),
            "mu_strategy": "adaptive",
            "hessian_approximation": "limited-memory",
            "limited_memory_max_history": 20,
            "limited_memory_initialization": "scalar2",
            "linear_solver": "mumps",
            "timing_statistics": "yes",
        }

    def solve(self, x0: ArrayLike) -> tuple[np.ndarray, dict]:
        """
        Solve the problem.

        Parameters
        ----------
        x0 : ArrayLike
            Initial guess for the decision variables.

        Returns
        -------
        result : dict
        """

        self.options.update(
            {
                "max_iter": self.max_iter,
                "max_cpu_time": float(self.max_time),
                "acceptable_tol": self.abs_obj_tol,
            }
        )

        x0 = np.asarray(x0)

        eval_jac_g_sparsity_indices = (np.array([]), np.array([]))
        eval_h_sparsity_indices = (np.array([]), np.array([]))

        def ipopt_derivative(x: NDArray[np.float64], out: NDArray[np.float64]):
            out[()] = self.gradient(x).astype(np.float64)
            return out

        # Set the optimization function
        nlp = Problem(
            n=x0.size,
            x_l=np.zeros_like(x0),
            x_u=np.inf * np.ones_like(x0),
            m=0,
            g_l=np.empty((0,)),
            g_u=np.empty((0,)),
            sparsity_indices_jac_g=eval_jac_g_sparsity_indices,
            sparsity_indices_h=eval_h_sparsity_indices,
            eval_f=self.objective,
            eval_grad_f=ipopt_derivative,
            eval_g=lambda _x, _out: None,
            eval_jac_g=lambda _x, _out: None,
            eval_h=None,
            ipopt_options=self.options,
        )

        x, _, status = nlp.solve(x0=x0)

        return x, status



import pytest
import numpy as np
import numpy as np

from pyRadPlan.core import PyRadPlanBaseModel


class DummyModel(PyRadPlanBaseModel):
    value: int
    array: list[int]
    nested: dict


@pytest.fixture
def dummy_instance():
    data = {
        "value": 10,
        "array": np.array([1, 2, 3]),
        "nested": {"a": {"a_1": np.array([1, 2, 3])}, "b": 2},
    }
    return DummyModel.model_validate(data)


@pytest.fixture
def another_dummy_instance():
    data = {
        "value": 10,
        "array": np.array([1, 2, 3]),
        "nested": {"a": {"a_1": np.array([1, 2, 3])}, "b": 2},
    }
    return DummyModel.model_validate(data)


@pytest.fixture
def different_dummy_instance():
    data = {
        "value": 10,
        "array": np.array([1, 2, 3]),
        "nested": {"a": {"a_1": np.array([1, 3, 2])}, "b": 2},
    }
    return DummyModel.model_validate(data)


def test_operator_equality(dummy_instance, another_dummy_instance):
    assert dummy_instance == another_dummy_instance


def test_operator_inequality(dummy_instance, different_dummy_instance):
    assert dummy_instance != different_dummy_instance



import pytest
import SimpleITK as sitk
import numpy as np
from typing import Union

# from pyRadPlan import plot_slice
from pyRadPlan.plan import Plan
from pyRadPlan.stf import SteeringInformation
from pyRadPlan.dij import Dij
from pyRadPlan.ct import CT
from pyRadPlan.cst import StructureSet
from pyRadPlan.machines import IonAccelerator
from pyRadPlan.core import Grid

from pyRadPlan.dose.engines import (
    DoseEngineBase,
)
from pyRadPlan.scenarios import NominalScenario


@pytest.fixture
def fake_dij(test_data_protons):
    dij_matRad = test_data_protons[4]
    dose_mat = np.ones((2000, 12))

    dij_matRad.physical_dose = np.empty((1,), dtype=object)
    dij_matRad.physical_dose[0] = dose_mat
    return dij_matRad


class DummyDoseEngine(DoseEngineBase):
    # short_name = "dummy"
    name = "Dummy Dose Engine"
    possible_radiation_modes = ["dummy"]

    def __init__(self, pln: Union[Plan, dict], fake_dij=[]):
        super().__init__(pln)
        self._fake_dij = fake_dij

    def _calc_dose(self, ct: CT, cst: StructureSet, stf: SteeringInformation) -> dict:
        return self._fake_dij


def test_DoseEngineBase_init(test_data_protons):
    dose_engine = DummyDoseEngine(pln=test_data_protons[0])

    assert isinstance(dose_engine, DoseEngineBase)
    assert dose_engine.name == "Dummy Dose Engine"
    assert dose_engine.is_dose_engine == True
    assert dose_engine.dose_grid["resolution"] == {"x": 10, "y": 10, "z": 10}
    assert dose_engine._ct_grid == None
    assert isinstance(dose_engine.mult_scen, NominalScenario)


def test_assign_properties_from_pln(test_data_protons):
    pln = test_data_protons[0]
    dose_engine = DummyDoseEngine(pln)

    assert dose_engine.dose_grid["resolution"] == {"x": 10, "y": 10, "z": 10}

    pln.prop_dose_calc["dose_grid"]["resolution"] = {"x": 20, "y": 15, "z": 10}
    dose_engine.assign_properties_from_pln(pln)

    assert dose_engine.name == "Dummy Dose Engine"
    assert dose_engine.is_dose_engine == True
    assert dose_engine.dose_grid["resolution"] == {"x": 20, "y": 15, "z": 10}


def test_calc_dose_forward(test_data_protons, fake_dij):
    pln, ct, cst, stf, _, _ = test_data_protons
    dose_engine = DummyDoseEngine(pln, fake_dij=fake_dij)
    result_py = dose_engine.calc_dose_forward(ct, cst, stf, w=None)

    assert isinstance(result_py["physical_dose"], sitk.Image)
    assert isinstance(result_py["let"], sitk.Image)

    result_py = sitk.GetArrayFromImage(result_py["physical_dose"])

    assert result_py.shape == (10, 20, 10)
    assert np.all(result_py == 12)


def test_calc_dose_influence(test_data_protons, fake_dij):
    pln, ct, cst, stf, _, _ = test_data_protons

    dose_engine = DummyDoseEngine(pln, fake_dij=fake_dij)
    result_py = dose_engine.calc_dose_influence(ct, cst, stf)

    assert isinstance(result_py, Dij)


def test_init_dose_calc(test_data_protons):
    pln, ct, cst, stf, _, _ = test_data_protons

    dose_engine = DummyDoseEngine(pln)
    dij = dose_engine._init_dose_calc(ct, cst, stf)

    assert dose_engine._num_of_columns_dij == 12

    assert len(dose_engine._vdose_grid) == 1152
    assert min(dose_engine._vdose_grid) == 211
    assert max(dose_engine._vdose_grid) == 1788

    assert len(dose_engine._vox_world_coords) == 1152
    assert np.array_equal(dose_engine._vox_world_coords[0], (-40.0, -90.0, -40.0))
    assert np.array_equal(dose_engine._vox_world_coords[-1], (30.0, 80.0, 30.0))
    assert np.array_equal(dose_engine._vox_world_coords, dose_engine._vox_world_coords_dose_grid)

    assert dose_engine._vdose_grid_mask.shape == (2000,)
    assert np.array_equal(dose_engine._vdose_grid_mask, dose_engine._vct_grid_mask)
    assert sum(dose_engine._vdose_grid_mask) == 1152

    assert isinstance(dose_engine._machine, IonAccelerator)
    assert isinstance(dose_engine._ct_grid, Grid)



"""Mean dose objective."""

from typing import Annotated
from pydantic import Field

from numba import njit
from numpy import ones

from ._objective import Objective, ParameterMetadata

# %% Class definition


class MeanDose(Objective):
    """
    Mean Dose objective.

    Attributes
    ----------
    d_ref : float
        referene mean dose to achieve

    Notes
    -----
    While we implement a reference value, we suggest to only use 0 as reference
    """

    name = "Mean Dose"

    d_ref: Annotated[float, Field(default=0.0, ge=0.0), ParameterMetadata(kind="reference")]

    def compute_objective(self, values):
        return _compute_objective(values, self.d_ref)

    def compute_gradient(self, values):
        return _compute_gradient(
            values,
            self.d_ref,
        )


@njit
def _compute_objective(dose, d_ref):
    return (dose.mean() - d_ref) ** 2


@njit
def _compute_gradient(dose, d_ref):
    grad = 2 * (dose.mean() - d_ref) * ones(dose.shape) / dose.size
    return grad



from typing import cast
import numpy as np
from ._base_pencilbeam_particle import ParticlePencilBeamEngineAbstract
from pyRadPlan.machines import IonPencilBeamKernel


class ParticleHongPencilBeamEngine(ParticlePencilBeamEngineAbstract):
    # constants
    short_name = "HongPB"
    name = "Hong Particle Pencil-Beam"
    possible_radiation_modes = ["protons", "helium", "carbon"]

    # private methods
    def _calc_particle_bixel(self, bixel):
        kernels = self._interpolate_kernels_in_depth(bixel)

        pb_kernel = cast(IonPencilBeamKernel, bixel["kernel"])

        # Lateral Component
        if self.lateral_model == "single":
            # Compute lateral sigma
            sigma_sq = kernels["sigma"] ** 2 + bixel["sigma_ini_sq"]
            lateral = np.exp(-bixel["radial_dist_sq"] / (2 * sigma_sq)) / (2 * np.pi * sigma_sq)
        elif self.lateral_model == "double":
            # Compute lateral sigmas
            sigma_sq_narrow = kernels["sigma_1"] ** 2 + bixel["sigma_ini_sq"]
            sigma_sq_broad = kernels["sigma_2"] ** 2 + bixel["sigma_ini_sq"]

            # Calculate lateral profile
            l_narr = np.exp(-bixel["radial_dist_sq"] / (2 * sigma_sq_narrow)) / (
                2 * np.pi * sigma_sq_narrow
            )
            l_bro = np.exp(-bixel["radial_dist_sq"] / (2 * sigma_sq_broad)) / (
                2 * np.pi * sigma_sq_broad
            )
            lateral = (1 - kernels["weight"]) * l_narr + kernels["weight"] * l_bro
        elif self.lateral_model == "multi":
            sigma_sq = kernels["sigma_multi"] ** 2 + bixel["sigma_ini_sq"]
            lateral = np.sum(
                (
                    np.column_stack(
                        (1 - np.sum(kernels["weight_multi"], axis=1), kernels["weight_multi"])
                    )
                    * np.exp(-bixel["radial_dist_sq"][:, np.newaxis] / (2 * sigma_sq))
                    / (2 * np.pi * sigma_sq)
                ),
                axis=1,
            )
        else:
            raise ValueError("Invalid Lateral Model")

        bixel["physical_dose"] = pb_kernel.lateral_cut_off.comp_fac * lateral * kernels["idd"]

        # Check if we have valid dose values
        if np.any(np.isnan(bixel["physical_dose"])) or np.any(bixel["physical_dose"] < 0):
            raise ValueError("Error in particle dose calculation.")

        if self.calc_let:
            bixel["let_dose"] = bixel["physical_dose"] * kernels["let"]

        if self.calc_bio_dose:
            # TODO: correct / adaptive alpha / beta values given tissue indices
            bixel_alpha = kernels["alpha"][0]
            bixel_beta = kernels["beta"][0]

            # Multiple with dose
            bixel["alpha_dose"] = bixel["physical_dose"] * bixel_alpha
            bixel["sqrt_beta_dose"] = bixel["physical_dose"] * np.sqrt(bixel_beta)

    @staticmethod
    def is_available(pln, machine):
        available, msg = [], []
        return available, msg
